#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "lcdi2c.h"

// RP2040 specific includes
#include "hardware/i2c.h"
#include "hardware/gpio.h"

// I2C instances
static i2c_inst_t* i2c_instances[2] = {i2c0, i2c1};

// Initialize I2C and GPIO pins for RP2040
static void lcdi2c_init_i2c(lcdi2c_i2cbus_obj_t *self) {
    // Validate I2C instance
    if (self->host > 1) {
        mp_raise_ValueError(MP_ERROR_TEXT("Invalid I2C host (must be 0 or 1)"));
    }
    
    // Get I2C instance
    i2c_inst_t *i2c = i2c_instances[self->host];
    self->i2c_inst = i2c;
    
    // Initialize I2C with the specified baudrate
    i2c_init(i2c, self->freq);
    
    // Configure GPIO pins for I2C
    gpio_set_function(self->sda_pin, GPIO_FUNC_I2C);
    gpio_set_function(self->scl_pin, GPIO_FUNC_I2C);
    
    // Enable pull-ups if requested
    if (self->sda_pullup) {
        gpio_pull_up(self->sda_pin);
    }
    
    if (self->scl_pullup) {
        gpio_pull_up(self->scl_pin);
    }
}

// Deinitialize I2C and GPIO
static void lcdi2c_deinit_i2c(lcdi2c_i2cbus_obj_t *self) {
    i2c_inst_t *i2c = (i2c_inst_t *)self->i2c_inst;
    
    if (i2c != NULL) {
        // Deinitialize I2C
        i2c_deinit(i2c);
        
        // Reset GPIO functions
        gpio_set_function(self->sda_pin, GPIO_FUNC_NULL);
        gpio_set_function(self->scl_pin, GPIO_FUNC_NULL);
    }
}

// Prepare control byte for I2C transaction
static uint8_t lcdi2c_prepare_control_byte(lcdi2c_i2cbus_obj_t *self, bool is_command) {
    if (self->disable_control_phase) {
        return 0; // No control byte used
    }
    
    uint8_t control_byte = 0;
    
    // Set or clear the DC bit based on whether this is a command or data
    bool dc_bit_value;
    if (self->dc_low_on_data) {
        dc_bit_value = is_command ? 1 : 0;
    } else {
        dc_bit_value = is_command ? 0 : 1;
    }
    
    if (dc_bit_value) {
        control_byte |= (1 << self->dc_bit_offset);
    }
    
    return control_byte;
}

// Send data over I2C with control byte(s)
static void lcdi2c_send_data(lcdi2c_i2cbus_obj_t *self, const uint8_t *data, size_t len, bool is_command) {
    i2c_inst_t *i2c = (i2c_inst_t *)self->i2c_inst;
    
    if (self->disable_control_phase) {
        // Send data directly without control byte
        int bytes_written = i2c_write_blocking(i2c, self->addr, data, len, false);
        if (bytes_written != len) {
            mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("I2C transmission failed"));
        }
        return;
    }
    
    // Prepare control byte
    uint8_t control_byte = lcdi2c_prepare_control_byte(self, is_command);
    
    // Allocate buffer for control byte(s) + data
    uint8_t *buffer = m_new(uint8_t, self->control_phase_bytes + len);
    if (buffer == NULL) {
        mp_raise_msg(&mp_type_MemoryError, MP_ERROR_TEXT("Failed to allocate I2C buffer"));
    }
    
    // Fill control bytes
    for (int i = 0; i < self->control_phase_bytes; i++) {
        buffer[i] = control_byte;
    }
    
    // Copy data
    memcpy(buffer + self->control_phase_bytes, data, len);
    
    // Send the complete buffer
    int bytes_written = i2c_write_blocking(i2c, self->addr, buffer, self->control_phase_bytes + len, false);
    if (bytes_written != (self->control_phase_bytes + len)) {
        m_del(uint8_t, buffer, self->control_phase_bytes + len);
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("I2C transmission failed"));
    }
    
    // Free the buffer
    m_del(uint8_t, buffer, self->control_phase_bytes + len);
}

// Receive data over I2C
static void lcdi2c_receive_data(lcdi2c_i2cbus_obj_t *self, uint8_t *data, size_t len, bool is_command) {
    i2c_inst_t *i2c = (i2c_inst_t *)self->i2c_inst;
    
    if (!self->disable_control_phase) {
        // Send control byte first
        uint8_t control_byte = lcdi2c_prepare_control_byte(self, is_command);
        
        int bytes_written = i2c_write_blocking(i2c, self->addr, &control_byte, 1, true);
        if (bytes_written != 1) {
            mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("I2C control byte transmission failed"));
        }
    }
    
    // Now receive the data
    int bytes_read = i2c_read_blocking(i2c, self->addr, data, len, false);
    if (bytes_read != len) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("I2C reception failed"));
    }
}