#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "lcdi8080_pio.h"

// RP2040 specific includes
#include "hardware/pio.h"
#include "hardware/dma.h"
#include "hardware/clocks.h"
#include "hardware/gpio.h"

// PIO program for 8-bit and 16-bit I8080 interface
// 8-bit mode:
// - Uses 8 consecutive GPIOs for data
// - WR pin is used as the base pin
// - DC pin controls command/data
// - CS pin for chip select

// PIO program for 8-bit data transfer
const uint16_t lcdi8080_pio_program_8bit_instructions[] = {
    // Pull data from FIFO to OSR (auto pull)
    // Set pins to output
    0x6001, // OUT PINS, 1    ; Output 1 bit to lowest pin (WR)
    0xe080, // SET PINS, 0    ; Set WR low (active)
    0x6008, // OUT PINS, 8    ; Output 8 bits of data
    0xe081, // SET PINS, 1    ; Set WR high (inactive)
};

// PIO program for 16-bit data transfer
const uint16_t lcdi8080_pio_program_16bit_instructions[] = {
    // Pull data from FIFO to OSR (auto pull)
    // Set pins to output
    0x6001, // OUT PINS, 1    ; Output 1 bit to lowest pin (WR)
    0xe080, // SET PINS, 0    ; Set WR low (active)
    0x6010, // OUT PINS, 16   ; Output 16 bits of data
    0xe081, // SET PINS, 1    ; Set WR high (inactive)
};

// Helper function declarations
static void lcdi8080_init_pio(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_deinit_pio(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_cs_select(lcdi8080_i80bus_obj_t *self, bool select);
static void lcdi8080_dc_set(lcdi8080_i80bus_obj_t *self, bool cmd, bool dummy, bool data);
static void lcdi8080_write_data(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len);
static void lcdi8080_read_data(lcdi8080_i80bus_obj_t *self, uint8_t *data, size_t len);

// Initialize PIO for 8080 parallel interface
static void lcdi8080_init_pio(lcdi8080_i80bus_obj_t *self) {
    // Choose which PIO instance to use (0 or 1)
    self->pio = pio0;
    
    // Find a free state machine on the PIO
    self->sm = pio_claim_unused_sm(self->pio, true);
    if (self->sm == -1) {
        self->pio = pio1;
        self->sm = pio_claim_unused_sm(self->pio, true);
        if (self->sm == -1) {
            mp_raise_msg(&mp_type_RuntimeError, MP_ERROR_TEXT("No free state machines"));
        }
    }
    
    // Get a free DMA channel
    self->dma_channel = dma_claim_unused_channel(true);
    if (self->dma_channel == -1) {
        mp_raise_msg(&mp_type_RuntimeError, MP_ERROR_TEXT("No free DMA channels"));
    }
    
    // Load the appropriate PIO program depending on data width
    if (self->data_width <= 8) {
        self->offset = pio_add_program(self->pio, lcdi8080_pio_program_8bit_instructions, 
                                      sizeof(lcdi8080_pio_program_8bit_instructions) / sizeof(uint16_t));
    } else {
        self->offset = pio_add_program(self->pio, lcdi8080_pio_program_16bit_instructions,
                                      sizeof(lcdi8080_pio_program_16bit_instructions) / sizeof(uint16_t));
    }
    
    // Configure state machine
    pio_sm_config config = pio_get_default_sm_config();
    
    // Configure autopull threshold based on data width
    sm_config_set_out_shift(&config, false, true, self->data_width + 1);
    
    // Configure pins
    sm_config_set_out_pins(&config, self->wr_pin, self->data_width + 1);
    
    // Set a sufficiently fast clock
    float div = (float)clock_get_hz(clk_sys) / (self->freq * 4); // 4 instructions per cycle
    sm_config_set_clkdiv(&config, div);
    
    // Set and activate sm
    pio_sm_init(self->pio, self->sm, self->offset, &config);
    
    // Initialize pins
    // WR pin (strobe)
    gpio_init(self->wr_pin);
    gpio_set_dir(self->wr_pin, GPIO_OUT);
    gpio_put(self->wr_pin, 1); // Inactive high
    
    // Data pins
    for (int i = 0; i < self->data_width; i++) {
        gpio_init(self->data_pins[i]);
        gpio_set_dir(self->data_pins[i], GPIO_OUT);
    }
    
    // DC pin
    gpio_init(self->dc_pin);
    gpio_set_dir(self->dc_pin, GPIO_OUT);
    gpio_put(self->dc_pin, self->dc_idle_high ? 1 : 0);
    
    // CS pin
    if (self->cs_pin >= 0) {
        gpio_init(self->cs_pin);
        gpio_set_dir(self->cs_pin, GPIO_OUT);
        gpio_put(self->cs_pin, self->cs_active_high ? 0 : 1); // Inactive
    }
    
    // Set pin function to PIO
    pio_gpio_init(self->pio, self->wr_pin);
    for (int i = 0; i < self->data_width; i++) {
        pio_gpio_init(self->pio, self->data_pins[i]);
    }
    
    // Start the state machine
    pio_sm_set_enabled(self->pio, self->sm, true);
}

// Deinitialize PIO
static void lcdi8080_deinit_pio(lcdi8080_i80bus_obj_t *self) {
    // Disable state machine
    pio_sm_set_enabled(self->pio, self->sm, false);
    
    // Unclaim resources
    pio_sm_unclaim(self->pio, self->sm);
    dma_channel_unclaim(self->dma_channel);
    
    // Return pins to normal GPIO
    gpio_set_function(self->wr_pin, GPIO_FUNC_SIO);
    for (int i = 0; i < self->data_width; i++) {
        gpio_set_function(self->data_pins[i], GPIO_FUNC_SIO);
    }
    
    // Free framebuffer if allocated
    if (self->framebuffer != NULL) {
        free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
}

// Set CS pin state
static void lcdi8080_cs_select(lcdi8080_i80bus_obj_t *self, bool select) {
    if (self->cs_pin >= 0) {
        gpio_put(self->cs_pin, (select == self->cs_active_high) ? 1 : 0);
    }
}

// Set DC pin state based on type (command, dummy, or data)
static void lcdi8080_dc_set(lcdi8080_i80bus_obj_t *self, bool cmd, bool dummy, bool data) {
    bool state;
    
    if (cmd) {
        state = self->dc_cmd_high;
    } else if (dummy) {
        state = self->dc_dummy_high;
    } else if (data) {
        state = self->dc_data_high;
    } else {
        state = self->dc_idle_high;
    }
    
    gpio_put(self->dc_pin, state ? 1 : 0);
}

// Write data using PIO
static void lcdi8080_write_data(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len) {
    // Wait for PIO FIFO to have space
    while (pio_sm_is_tx_fifo_full(self->pio, self->sm)) {
        tight_loop_contents();
    }
    
    if (self->data_width <= 8) {
        // 8-bit mode
        for (size_t i = 0; i < len; i++) {
            // Format: [data:8][wr:1]
            uint16_t value = ((uint16_t)data[i] << 1) | 0x01;
            pio_sm_put_blocking(self->pio, self->sm, value);
        }
    } else {
        // 16-bit mode
        const uint16_t *data16 = (const uint16_t *)data;
        size_t len16 = (len + 1) / 2; // Round up to handle odd bytes
        
        for (size_t i = 0; i < len16; i++) {
            uint16_t value = data16[i];
            
            // Handle byte swapping if needed
            if (self->swap_color_bytes) {
                value = ((value & 0xFF) << 8) | ((value >> 8) & 0xFF);
            }
            
            // Format: [data:16][wr:1]
            uint32_t pio_word = ((uint32_t)value << 1) | 0x01;
            pio_sm_put_blocking(self->pio, self->sm, pio_word);
        }
    }
}

// Read data function (Note: this requires bidirectional capabilities)
static void lcdi8080_read_data(lcdi8080_i80bus_obj_t *self, uint8_t *data, size_t len) {
    // This is more complex for PIO and would require reconfiguring the state machine for input
    // This implementation is simplified and incomplete
    
    // Reconfigure data pins as inputs
    for (int i = 0; i < self->data_width; i++) {
        gpio_set_dir(self->data_pins[i], GPIO_IN);
    }
    
    // Manual read process
    for (size_t i = 0; i < len; i++) {
        // Set WR high (inactive)
        gpio_put(self->wr_pin, 1);
        
        // Small delay
        for (volatile int j = 0; j < 10; j++) { }
        
        // Read data
        uint8_t byte = 0;
        for (int j = 0; j < 8; j++) {
            byte |= (gpio_get(self->data_pins[j]) ? 1 : 0) << j;
        }
        
        data[i] = byte;
        
        // Set WR low (active)
        gpio_put(self->wr_pin, 0);
        
        // Small delay
        for (volatile int j = 0; j < 10; j++) { }
    }
    
    // Reconfigure data pins as outputs
    for (int i = 0; i < self->data_width; i++) {
        gpio_set_dir(self->data_pins[i], GPIO_OUT);
    }
}

// Export functions to header
void lcdi8080_pio_dma_handler(void);
void lcdi8080_pio_dma_configure(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len);

// Configure DMA transfer
void lcdi8080_pio_dma_configure(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len) {
    // Configure the DMA channel
    dma_channel_config c = dma_channel_get_default_config(self->dma_channel);
    
    if (self->data_width <= 8) {
        // 8-bit mode: transfer 16-bit values to PIO (8-bit data + 1-bit WR)
        channel_config_set_transfer_data_size(&c, DMA_SIZE_16);
        
        // Create a temporary buffer with the proper format for the PIO
        uint16_t *buffer = malloc(len * sizeof(uint16_t));
        if (buffer == NULL) {
            mp_raise_msg(&mp_type_MemoryError, MP_ERROR_TEXT("Failed to allocate DMA buffer"));
            return;
        }
        
        // Format data for the PIO: [data:8][wr:1]
        for (size_t i = 0; i < len; i++) {
            buffer[i] = ((uint16_t)data[i] << 1) | 0x01;
        }
        
        // Configure DMA
        channel_config_set_read_increment(&c, true);
        channel_config_set_write_increment(&c, false);
        channel_config_set_dreq(&c, pio_get_dreq(self->pio, self->sm, true));
        
        // Set up the DMA channel
        dma_channel_configure(
            self->dma_channel,
            &c,
            &self->pio->txf[self->sm],  // Write to PIO TX FIFO
            buffer,                      // Read from our buffer
            len,                         // Number of transfers
            true                         // Start immediately
        );
        
        // Wait for DMA to complete
        dma_channel_wait_for_finish_blocking(self->dma_channel);
        
        // Free the temporary buffer
        free(buffer);
    } else {
        // 16-bit mode: transfer 32-bit values to PIO (16-bit data + 1-bit WR)
        channel_config_set_transfer_data_size(&c, DMA_SIZE_32);
        
        // Create a temporary buffer with the proper format for the PIO
        uint32_t *buffer = malloc((len / 2) * sizeof(uint32_t));
        if (buffer == NULL) {
            mp_raise_msg(&mp_type_MemoryError, MP_ERROR_TEXT("Failed to allocate DMA buffer"));
            return;
        }
        
        // Format data for the PIO: [data:16][wr:1]
        const uint16_t *data16 = (const uint16_t *)data;
        size_t len16 = len / 2;
        
        for (size_t i = 0; i < len16; i++) {
            uint16_t value = data16[i];
            
            // Handle byte swapping if needed
            if (self->swap_color_bytes) {
                value = ((value & 0xFF) << 8) | ((value >> 8) & 0xFF);
            }
            
            buffer[i] = ((uint32_t)value << 1) | 0x01;
        }
        
        // Configure DMA
        channel_config_set_read_increment(&c, true);
        channel_config_set_write_increment(&c, false);
        channel_config_set_dreq(&c, pio_get_dreq(self->pio, self->sm, true));
        
        // Set up the DMA channel
        dma_channel_configure(
            self->dma_channel,
            &c,
            &self->pio->txf[self->sm],  // Write to PIO TX FIFO
            buffer,                      // Read from our buffer
            len16,                       // Number of transfers
            true                         // Start immediately
        );
        
        // Wait for DMA to complete
        dma_channel_wait_for_finish_blocking(self->dma_channel);
        
        // Free the temporary buffer
        free(buffer);
    }
}