#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "py/mperrno.h"
#include "py/objarray.h"

#include "lcdi8080_pio.h"

// RP2040 specific includes
#include "hardware/pio.h"
#include "hardware/dma.h"
#include "hardware/gpio.h"

// Function declarations from lcdi8080_pio.c
void lcdi8080_pio_dma_configure(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len);

// Helper function declarations
static void lcdi8080_init_pio(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_deinit_pio(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_cs_select(lcdi8080_i80bus_obj_t *self, bool select);
static void lcdi8080_dc_set(lcdi8080_i80bus_obj_t *self, bool cmd, bool dummy, bool data);
static void lcdi8080_write_data(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len);
static void lcdi8080_read_data(lcdi8080_i80bus_obj_t *self, uint8_t *data, size_t len);

// I80Bus constructor
STATIC mp_obj_t lcdi8080_i80bus_make_new(const mp_obj_type_t *type, size_t n_args, size_t n_kw, const mp_obj_t *all_args) {
    // Define kwargs for I80Bus constructor
    enum { 
        ARG_dc, ARG_wr, ARG_data0, ARG_data1, ARG_data2, ARG_data3, ARG_data4, ARG_data5, 
        ARG_data6, ARG_data7, ARG_data8, ARG_data9, ARG_data10, ARG_data11, ARG_data12, 
        ARG_data13, ARG_data14, ARG_data15, ARG_cs, ARG_freq, ARG_dc_idle_high, ARG_dc_cmd_high,
        ARG_dc_dummy_high, ARG_dc_data_high, ARG_cs_active_high, ARG_reverse_color_bits,
        ARG_swap_color_bytes
    };
    
    static const mp_arg_t allowed_args[] = {
        { MP_QSTR_dc, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_wr, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data0, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data1, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data2, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data3, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data4, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data5, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data6, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data7, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data8, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data9, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data10, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data11, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data12, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data13, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data14, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data15, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_cs, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_freq, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 10000000} },
        { MP_QSTR_dc_idle_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_cmd_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_dummy_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_data_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = true} },
        { MP_QSTR_cs_active_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_reverse_color_bits, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_swap_color_bytes, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
    };
    
    mp_arg_val_t args[MP_ARRAY_SIZE(allowed_args)];
    mp_arg_parse_all_kw_array(n_args, n_kw, all_args, MP_ARRAY_SIZE(allowed_args), allowed_args, args);
    
    // Create new I80Bus object
    lcdi8080_i80bus_obj_t *self = mp_obj_malloc(lcdi8080_i80bus_obj_t, &lcdi8080_i80bus_type);
    
    // Set pins and configuration
    self->dc_pin = args[ARG_dc].u_int;
    self->wr_pin = args[ARG_wr].u_int;
    self->cs_pin = args[ARG_cs].u_int;
    
    // Set data pins
    self->data_pins[0] = args[ARG_data0].u_int;
    self->data_pins[1] = args[ARG_data1].u_int;
    self->data_pins[2] = args[ARG_data2].u_int;
    self->data_pins[3] = args[ARG_data3].u_int;
    self->data_pins[4] = args[ARG_data4].u_int;
    self->data_pins[5] = args[ARG_data5].u_int;
    self->data_pins[6] = args[ARG_data6].u_int;
    self->data_pins[7] = args[ARG_data7].u_int;
    
    // Determine bus width (8-bit or 16-bit)
    self->data_width = 8;  // Default to 8-bit
    
    // Check if any 16-bit data pins are defined
    if (args[ARG_data8].u_int != -1) {
        self->data_pins[8] = args[ARG_data8].u_int;
        self->data_width = 16;
    }
    if (args[ARG_data9].u_int != -1) {
        self->data_pins[9] = args[ARG_data9].u_int;
    }
    if (args[ARG_data10].u_int != -1) {
        self->data_pins[10] = args[ARG_data10].u_int;
    }
    if (args[ARG_data11].u_int != -1) {
        self->data_pins[11] = args[ARG_data11].u_int;
    }
    if (args[ARG_data12].u_int != -1) {
        self->data_pins[12] = args[ARG_data12].u_int;
    }
    if (args[ARG_data13].u_int != -1) {
        self->data_pins[13] = args[ARG_data13].u_int;
    }
    if (args[ARG_data14].u_int != -1) {
        self->data_pins[14] = args[ARG_data14].u_int;
    }
    if (args[ARG_data15].u_int != -1) {
        self->data_pins[15] = args[ARG_data15].u_int;
    }
    
    // Set other configuration parameters
    self->freq = args[ARG_freq].u_int;
    self->dc_idle_high = args[ARG_dc_idle_high].u_bool;
    self->dc_cmd_high = args[ARG_dc_cmd_high].u_bool;
    self->dc_dummy_high = args[ARG_dc_dummy_high].u_bool;
    self->dc_data_high = args[ARG_dc_data_high].u_bool;
    self->cs_active_high = args[ARG_cs_active_high].u_bool;
    self->reverse_color_bits = args[ARG_reverse_color_bits].u_bool;
    self->swap_color_bytes = args[ARG_swap_color_bytes].u_bool;
    
    // Initialize framebuffer to NULL
    self->framebuffer = NULL;
    self->framebuffer_size = 0;
    
    // Set callback to None
    self->callback = mp_const_none;
    
    //// filepath: c:\Users\jgpei\Desktop\py\micropython_drivers\rp2040\lcdi8080_pio\modlcdi8080_pio.c
#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "py/mperrno.h"
#include "py/objarray.h"

#include "lcdi8080_pio.h"

// RP2040 specific includes
#include "hardware/pio.h"
#include "hardware/dma.h"
#include "hardware/gpio.h"

// Function declarations from lcdi8080_pio.c
void lcdi8080_pio_dma_configure(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len);

// Helper function declarations
static void lcdi8080_init_pio(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_deinit_pio(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_cs_select(lcdi8080_i80bus_obj_t *self, bool select);
static void lcdi8080_dc_set(lcdi8080_i80bus_obj_t *self, bool cmd, bool dummy, bool data);
static void lcdi8080_write_data(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len);
static void lcdi8080_read_data(lcdi8080_i80bus_obj_t *self, uint8_t *data, size_t len);

// I80Bus constructor
STATIC mp_obj_t lcdi8080_i80bus_make_new(const mp_obj_type_t *type, size_t n_args, size_t n_kw, const mp_obj_t *all_args) {
    // Define kwargs for I80Bus constructor
    enum { 
        ARG_dc, ARG_wr, ARG_data0, ARG_data1, ARG_data2, ARG_data3, ARG_data4, ARG_data5, 
        ARG_data6, ARG_data7, ARG_data8, ARG_data9, ARG_data10, ARG_data11, ARG_data12, 
        ARG_data13, ARG_data14, ARG_data15, ARG_cs, ARG_freq, ARG_dc_idle_high, ARG_dc_cmd_high,
        ARG_dc_dummy_high, ARG_dc_data_high, ARG_cs_active_high, ARG_reverse_color_bits,
        ARG_swap_color_bytes
    };
    
    static const mp_arg_t allowed_args[] = {
        { MP_QSTR_dc, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_wr, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data0, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data1, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data2, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data3, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data4, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data5, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data6, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data7, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data8, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data9, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data10, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data11, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data12, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data13, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data14, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data15, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_cs, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_freq, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 10000000} },
        { MP_QSTR_dc_idle_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_cmd_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_dummy_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_data_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = true} },
        { MP_QSTR_cs_active_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_reverse_color_bits, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_swap_color_bytes, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
    };
    
    mp_arg_val_t args[MP_ARRAY_SIZE(allowed_args)];
    mp_arg_parse_all_kw_array(n_args, n_kw, all_args, MP_ARRAY_SIZE(allowed_args), allowed_args, args);
    
    // Create new I80Bus object
    lcdi8080_i80bus_obj_t *self = mp_obj_malloc(lcdi8080_i80bus_obj_t, &lcdi8080_i80bus_type);
    
    // Set pins and configuration
    self->dc_pin = args[ARG_dc].u_int;
    self->wr_pin = args[ARG_wr].u_int;
    self->cs_pin = args[ARG_cs].u_int;
    
    // Set data pins
    self->data_pins[0] = args[ARG_data0].u_int;
    self->data_pins[1] = args[ARG_data1].u_int;
    self->data_pins[2] = args[ARG_data2].u_int;
    self->data_pins[3] = args[ARG_data3].u_int;
    self->data_pins[4] = args[ARG_data4].u_int;
    self->data_pins[5] = args[ARG_data5].u_int;
    self->data_pins[6] = args[ARG_data6].u_int;
    self->data_pins[7] = args[ARG_data7].u_int;
    
    // Determine bus width (8-bit or 16-bit)
    self->data_width = 8;  // Default to 8-bit
    
    // Check if any 16-bit data pins are defined
    if (args[ARG_data8].u_int != -1) {
        self->data_pins[8] = args[ARG_data8].u_int;
        self->data_width = 16;
    }
    if (args[ARG_data9].u_int != -1) {
        self->data_pins[9] = args[ARG_data9].u_int;
    }
    if (args[ARG_data10].u_int != -1) {
        self->data_pins[10] = args[ARG_data10].u_int;
    }
    if (args[ARG_data11].u_int != -1) {
        self->data_pins[11] = args[ARG_data11].u_int;
    }
    if (args[ARG_data12].u_int != -1) {
        self->data_pins[12] = args[ARG_data12].u_int;
    }
    if (args[ARG_data13].u_int != -1) {
        self->data_pins[13] = args[ARG_data13].u_int;
    }
    if (args[ARG_data14].u_int != -1) {
        self->data_pins[14] = args[ARG_data14].u_int;
    }
    if (args[ARG_data15].u_int != -1) {
        self->data_pins[15] = args[ARG_data15].u_int;
    }
    
    // Set other configuration parameters
    self->freq = args[ARG_freq].u_int;
    self->dc_idle_high = args[ARG_dc_idle_high].u_bool;
    self->dc_cmd_high = args[ARG_dc_cmd_high].u_bool;
    self->dc_dummy_high = args[ARG_dc_dummy_high].u_bool;
    self->dc_data_high = args[ARG_dc_data_high].u_bool;
    self->cs_active_high = args[ARG_cs_active_high].u_bool;
    self->reverse_color_bits = args[ARG_reverse_color_bits].u_bool;
    self->swap_color_bytes = args[ARG_swap_color_bytes].u_bool;
    
    // Initialize framebuffer to NULL
    self->framebuffer = NULL;
    self->framebuffer_size = 0;
    
    // Set callback to None
    self->callback = mp_const_none;
    
    // Initialize the PIO hardware
    lcdi8080_init_pio(self);
    
    return MP_OBJ_FROM_PTR(self);
}

// init(width, height, bpp, buffer_size, rgb565_byte_swap, cmd_bits, param_bits)
STATIC mp_obj_t lcdi8080_i80bus_init(mp_obj_t self_in, mp_obj_t width_in, mp_obj_t height_in, 
                                     mp_obj_t bpp_in, mp_obj_t buffer_size_in, 
                                     mp_obj_t rgb565_byte_swap_in, mp_obj_t cmd_bits_in, 
                                     mp_obj_t param_bits_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Set display parameters
    self->width = mp_obj_get_int(width_in);
    self->height = mp_obj_get_int(height_in);
    self->bpp = mp_obj_get_int(bpp_in);
    self->buffer_size = mp_obj_get_int(buffer_size_in);
    self->rgb565_byte_swap = mp_obj_is_true(rgb565_byte_swap_in);
    self->cmd_bits = mp_obj_get_int(cmd_bits_in);
    self->param_bits = mp_obj_get_int(param_bits_in);
    
    return mp_const_none;
}
MP_DEFINE_CONST_FUN_OBJ_8(lcdi8080_i80bus_init_obj, lcdi8080_i80bus_init);

// deinit()
STATIC mp_obj_t lcdi8080_i80bus_deinit(mp_obj_t self_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Deinitialize the PIO hardware
    lcdi8080_deinit_pio(self);
    
    return mp_const_none;
}
MP_DEFINE_CONST_FUN_OBJ_1(lcdi8080_i80bus_deinit_obj, lcdi8080_i80bus_deinit);

// register_callback(callback)
STATIC mp_obj_t lcdi8080_i80bus_register_callback(mp_obj_t self_in, mp_obj_t callback_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Store the callback function
    if (callback_in == mp_const_none) {
        self->callback = mp_const_none;
    } else if (mp_obj_is_callable(callback_in)) {
        self->callback = callback_in;
    } else {
        mp_raise_ValueError(MP_ERROR_TEXT("callback must be callable"));
    }
    
    return mp_const_none;
}
MP_DEFINE_CONST_FUN_OBJ_2(lcdi8080_i80bus_register_callback_obj, lcdi8080_i80bus_register_callback);

// tx_color(cmd, data, start_x, start_y, end_x, end_y, rotation, last_update)
STATIC mp_obj_t lcdi8080_i80bus_tx_color(size_t n_args, const mp_obj_t *args) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(args[2], &bufinfo, MP_BUFFER_READ);
    int start_x = mp_obj_get_int(args[3]);
    int start_y = mp_obj_get_int(args[4]);
    int end_x = mp_obj_get_int(args[5]);
    int end_y = mp_obj_get_int(args[6]);
    int rotation = mp_obj_get_int(args[7]);
    bool last_update = mp_obj_is_true(args[8]);
    
    // Select the device
    lcdi8080_cs_select(self, true);
    
    // Set DC to command mode
    lcdi8080_dc_set(self, true, false, false);
    
    // Send command
    uint8_t cmd_buf[1] = {cmd};
    lcdi8080_write_data(self, cmd_buf, 1);
    
    // Set DC to data mode
    lcdi8080_dc_set(self, false, false, true);
    
    // Send color data using DMA for efficiency
    lcdi8080_pio_dma_configure(self, bufinfo.buf, bufinfo.len);
    
    // If there's a callback and this is the last update, call it
    if (self->callback != mp_const_none && last_update) {
        mp_obj_t args[3] = {
            self->callback,
            mp_obj_new_int(bufinfo.len),
            MP_OBJ_FROM_PTR(self)
        };
        mp_call_method_n_kw(1, 0, args);
    }
    
    // Deselect the device
    lcdi8080_cs_select(self, false);
    
    return mp_const_none;
}
MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdi8080_i80bus_tx_color_obj, 9, 9, lcdi8080_i80bus_tx_color);

// tx_param(cmd, data)
STATIC mp_obj_t lcdi8080_i80bus_tx_param(mp_obj_t self_in, mp_obj_t cmd_in, mp_obj_t data_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    uint8_t cmd = mp_obj_get_int(cmd_in);
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(data_in, &bufinfo, MP_BUFFER_READ);
    
    // Select the device
    lcdi8080_cs_select(self, true);
    
    // Set DC to command mode
    lcdi8080_dc_set(self, true, false, false);
    
    // Send command
    uint8_t cmd_buf[1] = {cmd};
    lcdi8080_write_data(self, cmd_buf, 1);
    
    // Set DC to data mode
    lcdi8080_dc_set(self, false, false, true);
    
    // Send parameter data
    lcdi8080_write_data(self, bufinfo.buf, bufinfo.len);
    
    // Deselect the device
    lcdi8080_cs_select(self, false);
    
    return mp_const_none;
}
MP_DEFINE_CONST_FUN_OBJ_3(lcdi8080_i80bus_tx_param_obj, lcdi8080_i80bus_tx_param);

// rx_param(cmd, params)
STATIC mp_obj_t lcdi8080_i80bus_rx_param(mp_obj_t self_in, mp_obj_t cmd_in, mp_obj_t params_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    uint8_t cmd = mp_obj_get_int(cmd_in);
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(params_in, &bufinfo, MP_BUFFER_WRITE);
    
    // Select the device
    lcdi8080_cs_select(self, true);
    
    // Set DC to command mode
    lcdi8080_dc_set(self, true, false, false);
    
    // Send command
    uint8_t cmd_buf[1] = {cmd};
    lcdi8080_write_data(self, cmd_buf, 1);
    
    // Set DC to data mode
    lcdi8080_dc_set(self, false, false, true);
    
    // Read parameter data
    lcdi8080_read_data(self, bufinfo.buf, bufinfo.len);
    
    // Deselect the device
    lcdi8080_cs_select(self, false);
    
    return mp_const_none;
}
MP_DEFINE_CONST_FUN_OBJ_3(lcdi8080_i80bus_rx_param_obj, lcdi8080_i80bus_rx_param);

// get_lane_count()
STATIC mp_obj_t lcdi8080_i80bus_get_lane_count(mp_obj_t self_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Return the number of data lanes (bus width)
    return mp_obj_new_int(self->data_width);
}
MP_DEFINE_CONST_FUN_OBJ_1(lcdi8080_i80bus_get_lane_count_obj, lcdi8080_i80bus_get_lane_count);

// allocate_framebuffer(size, caps)
STATIC mp_obj_t lcdi8080_i80bus_allocate_framebuffer(mp_obj_t self_in, mp_obj_t size_in, mp_obj_t caps_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    int size = mp_obj_get_int(size_in);
    // int caps = mp_obj_get_int(caps_in); // Currently not used
    
    // Free existing framebuffer if it exists
    if (self->framebuffer != NULL) {
        free(self->framebuffer);
        self->framebuffer = NULL;
    }
    
    // Allocate new framebuffer
    self->framebuffer = malloc(size);
    if (self->framebuffer == NULL) {
        // Failed to allocate
        return mp_const_none;
    }
    
    self->framebuffer_size = size;
    
    // Return a memoryview of the framebuffer
    mp_obj_array_t *view = m_new_obj(mp_obj_array_t);
    mp_obj_memoryview_init(view, 'B', 0, size, self->framebuffer);
    
    return MP_OBJ_FROM_PTR(view);
}
MP_DEFINE_CONST_FUN_OBJ_3(lcdi8080_i80bus_allocate_framebuffer_obj, lcdi8080_i80bus_allocate_framebuffer);

// free_framebuffer(framebuffer)
STATIC mp_obj_t lcdi8080_i80bus_free_framebuffer(mp_obj_t self_in, mp_obj_t framebuffer_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Free the framebuffer memory
    if (self->framebuffer != NULL) {
        free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    return mp_const_none;
}
MP_DEFINE_CONST_FUN_OBJ_2(lcdi8080_i80bus_free_framebuffer_obj, lcdi8080_i80bus_free_framebuffer);

// Dictionary of methods
STATIC const mp_rom_map_elem_t lcdi8080_i80bus_locals_dict_table[] = {
    { MP_ROM_QSTR(MP_QSTR_init), MP_ROM_PTR(&lcdi8080_i80bus_init_obj) },
    { MP_ROM_QSTR(MP_QSTR_deinit), MP_ROM_PTR(&lcdi8080_i80bus_deinit_obj) },
    { MP_ROM_QSTR(MP_QSTR_register_callback), MP_ROM_PTR(&lcdi8080_i80bus_register_callback_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_color), MP_ROM_PTR(&lcdi8080_i80bus_tx_color_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_param), MP_ROM_PTR(&lcdi8080_i80bus_tx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_rx_param), MP_ROM_PTR(&lcdi8080_i80bus_rx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_get_lane_count), MP_ROM_PTR(&lcdi8080_i80bus_get_lane_count_obj) },
    { MP_ROM_QSTR(MP_QSTR_allocate_framebuffer), MP_ROM_PTR(&lcdi8080_i80bus_allocate_framebuffer_obj) },
    { MP_ROM_QSTR(MP_QSTR_free_framebuffer), MP_ROM_PTR(&lcdi8080_i80bus_free_framebuffer_obj) },
};
STATIC MP_DEFINE_CONST_DICT(lcdi8080_i80bus_locals_dict, lcdi8080_i80bus_locals_dict_table);

// Define the I80Bus type
const mp_obj_type_t lcdi8080_i80bus_type = {
    { &mp_type_type },
    .name = MP_QSTR_I80Bus,
    .make_new = lcdi8080_i80bus_make_new,
    .locals_dict = (mp_obj_dict_t*)&lcdi8080_i80bus_locals_dict,
};

// Module initialization
STATIC const mp_rom_map_elem_t lcdi8080_module_globals_table[] = {
    { MP_ROM_QSTR(MP_QSTR___name__), MP_ROM_QSTR(MP_QSTR_lcdi8080) },
    { MP_ROM_QSTR(MP_QSTR_I80Bus), MP_ROM_PTR(&lcdi8080_i80bus_type) },
};
STATIC MP_DEFINE_CONST_DICT(lcdi8080_module_globals, lcdi8080_module_globals_table);

// Define the module
const mp_obj_module_t lcdi8080_module = {
    .base = { &mp_type_module },
    .globals = (mp_obj_dict_t*)&lcdi8080_module_globals,
};

// Register the module
MP_REGISTER_MODULE(MP_QSTR_lcdi8080, lcdi8080_module);