"""
MicroPython ILI9341 Display Demo using Intel 8080 Parallel Interface
For RP2040 boards (Raspberry Pi Pico)

Wiring:
- DC: GPIO 8
- WR: GPIO 9
- CS: GPIO 10
- D0: GPIO 0
- D1: GPIO 1
- D2: GPIO 2
- D3: GPIO 3
- D4: GPIO 4
- D5: GPIO 5
- D6: GPIO 6
- D7: GPIO 7
- RST: GPIO 11 (Reset pin, controlled separately)

This demo initializes the ILI9341 display and draws some basic graphics.
"""

import machine
import time
import lcdi8080
import framebuf

# Display dimensions
WIDTH = 240
HEIGHT = 320
BPP = 16  # RGB565

# Reset pin
rst_pin = machine.Pin(11, machine.Pin.OUT)

# Define colors (RGB565 format)
BLACK = 0x0000
BLUE = 0x001F
RED = 0xF800
GREEN = 0x07E0
CYAN = 0x07FF
MAGENTA = 0xF81F
YELLOW = 0xFFE0
WHITE = 0xFFFF

# ILI9341 command constants
ILI9341_NOP = 0x00
ILI9341_SWRESET = 0x01
ILI9341_RDDID = 0x04
ILI9341_RDDST = 0x09
ILI9341_SLPIN = 0x10
ILI9341_SLPOUT = 0x11
ILI9341_PTLON = 0x12
ILI9341_NORON = 0x13
ILI9341_RDMODE = 0x0A
ILI9341_RDMADCTL = 0x0B
ILI9341_RDPIXFMT = 0x0C
ILI9341_RDIMGFMT = 0x0D
ILI9341_RDSELFDIAG = 0x0F
ILI9341_INVOFF = 0x20
ILI9341_INVON = 0x21
ILI9341_GAMMASET = 0x26
ILI9341_DISPOFF = 0x28
ILI9341_DISPON = 0x29
ILI9341_CASET = 0x2A
ILI9341_PASET = 0x2B
ILI9341_RAMWR = 0x2C
ILI9341_RAMRD = 0x2E
ILI9341_PTLAR = 0x30
ILI9341_MADCTL = 0x36
ILI9341_PIXFMT = 0x3A
ILI9341_FRMCTR1 = 0xB1
ILI9341_FRMCTR2 = 0xB2
ILI9341_FRMCTR3 = 0xB3
ILI9341_INVCTR = 0xB4
ILI9341_DFUNCTR = 0xB6
ILI9341_PWCTR1 = 0xC0
ILI9341_PWCTR2 = 0xC1
ILI9341_PWCTR3 = 0xC2
ILI9341_PWCTR4 = 0xC3
ILI9341_PWCTR5 = 0xC4
ILI9341_VMCTR1 = 0xC5
ILI9341_VMCTR2 = 0xC7
ILI9341_RDID1 = 0xDA
ILI9341_RDID2 = 0xDB
ILI9341_RDID3 = 0xDC
ILI9341_RDID4 = 0xDD
ILI9341_GMCTRP1 = 0xE0
ILI9341_GMCTRN1 = 0xE1

# MADCTL rotation values
MADCTL_MX = 0x40
MADCTL_MY = 0x80
MADCTL_MV = 0x20
MADCTL_ML = 0x10
MADCTL_BGR = 0x08
MADCTL_MH = 0x04


class ILI9341:
    def __init__(self):
        # Create 8080 bus interface
        self.bus = lcdi8080.I80Bus(
            dc=8,
            wr=9,
            data0=0,
            data1=1,
            data2=2,
            data3=3,
            data4=4,
            data5=5,
            data6=6,
            data7=7,
            cs=10,
            freq=10000000,  # 10MHz
            dc_idle_high=False,
            dc_cmd_high=False,
            dc_data_high=True
        )
        
        # Initialize the bus with display parameters
        self.bus.init(WIDTH, HEIGHT, BPP, 32 * 1024, False, 8, 8)
        
        # Create and allocate framebuffer
        self.buffer_size = WIDTH * HEIGHT * 2  # 2 bytes per pixel in RGB565
        self.fb_mem = bytearray(self.buffer_size)
        self.fb = framebuf.FrameBuffer(self.fb_mem, WIDTH, HEIGHT, framebuf.RGB565)
        
        # Reset and initialize display
        self.reset()
        self.init_display()
        
        # Default rotation
        self.rotation = 0
        self.set_rotation(0)

    def reset(self):
        """Hardware reset the display"""
        rst_pin.value(1)
        time.sleep_ms(50)
        rst_pin.value(0)
        time.sleep_ms(50)
        rst_pin.value(1)
        time.sleep_ms(150)

    def init_display(self):
        """Initialize display with recommended settings for ILI9341"""
        # Exit sleep mode
        self.bus.tx_param(ILI9341_SLPOUT, bytearray())
        time.sleep_ms(120)
        
        # Set pixel format to RGB565 (16 bits per pixel)
        self.bus.tx_param(ILI9341_PIXFMT, bytearray([0x55]))
        
        # Frame control
        self.bus.tx_param(ILI9341_FRMCTR1, bytearray([0x00, 0x18]))
        
        # Display function control
        self.bus.tx_param(ILI9341_DFUNCTR, bytearray([0x08, 0x82, 0x27]))
        
        # Power control settings
        self.bus.tx_param(ILI9341_PWCTR1, bytearray([0x23]))
        self.bus.tx_param(ILI9341_PWCTR2, bytearray([0x10]))
        self.bus.tx_param(ILI9341_VMCTR1, bytearray([0x3e, 0x28]))
        self.bus.tx_param(ILI9341_VMCTR2, bytearray([0x86]))
        
        # Set gamma curve
        self.bus.tx_param(ILI9341_GAMMASET, bytearray([0x01]))
        
        # Positive Gamma Correction
        self.bus.tx_param(ILI9341_GMCTRP1, bytearray([
            0x0F, 0x31, 0x2B, 0x0C, 0x0E, 0x08, 0x4E, 0xF1,
            0x37, 0x07, 0x10, 0x03, 0x0E, 0x09, 0x00
        ]))
        
        # Negative Gamma Correction
        self.bus.tx_param(ILI9341_GMCTRN1, bytearray([
            0x00, 0x0E, 0x14, 0x03, 0x11, 0x07, 0x31, 0xC1,
            0x48, 0x08, 0x0F, 0x0C, 0x31, 0x36, 0x0F
        ]))
        
        # Normal display mode
        self.bus.tx_param(ILI9341_NORON, bytearray())
        
        # Turn display on
        self.bus.tx_param(ILI9341_DISPON, bytearray())
        time.sleep_ms(20)

    def set_rotation(self, rotation):
        """Set display rotation (0-3)"""
        self.rotation = rotation % 4
        
        # MADCTL values for different rotations
        if self.rotation == 0:  # Portrait
            madctl = MADCTL_BGR
        elif self.rotation == 1:  # Landscape
            madctl = MADCTL_MV | MADCTL_BGR
        elif self.rotation == 2:  # Portrait flipped
            madctl = MADCTL_MY | MADCTL_BGR
        else:  # Landscape flipped
            madctl = MADCTL_MX | MADCTL_MY | MADCTL_MV | MADCTL_BGR
        
        self.bus.tx_param(ILI9341_MADCTL, bytearray([madctl]))

    def set_window(self, x0, y0, x1, y1):
        """Set active drawing area (column and page address)"""
        # Column address set
        self.bus.tx_param(ILI9341_CASET, bytearray([
            (x0 >> 8) & 0xFF, x0 & 0xFF, 
            (x1 >> 8) & 0xFF, x1 & 0xFF
        ]))
        
        # Page address set
        self.bus.tx_param(ILI9341_PASET, bytearray([
            (y0 >> 8) & 0xFF, y0 & 0xFF, 
            (y1 >> 8) & 0xFF, y1 & 0xFF
        ]))

    def display(self):
        """Update the display with contents of the framebuffer"""
        # Set the drawing window to full screen
        self.set_window(0, 0, WIDTH-1, HEIGHT-1)
        
        # Send framebuffer data
        self.bus.tx_color(
            ILI9341_RAMWR, 
            self.fb_mem, 
            0, 0, WIDTH-1, HEIGHT-1, 
            self.rotation, True
        )

    def fill(self, color):
        """Fill entire display with a color"""
        self.fb.fill(color)

    def pixel(self, x, y, color):
        """Set a pixel color"""
        self.fb.pixel(x, y, color)

    def line(self, x0, y0, x1, y1, color):
        """Draw a line"""
        self.fb.line(x0, y0, x1, y1, color)

    def rect(self, x, y, w, h, color, fill=False):
        """Draw a rectangle"""
        if fill:
            self.fb.fill_rect(x, y, w, h, color)
        else:
            self.fb.rect(x, y, w, h, color)

    def circle(self, x, y, r, color, fill=False):
        """Draw a circle"""
        for i in range(-r, r+1):
            for j in range(-r, r+1):
                if i*i + j*j <= r*r:
                    if fill or i*i + j*j >= (r-1)**2:
                        self.fb.pixel(x+i, y+j, color)

    def text(self, text, x, y, color):
        """Display text"""
        self.fb.text(text, x, y, color)


def demo():
    """Run a demo of the display capabilities"""
    display = ILI9341()
    
    # Clear screen
    display.fill(BLACK)
    display.display()
    time.sleep(1)
    
    # Draw some shapes
    display.rect(10, 10, 100, 50, RED, fill=True)
    display.rect(120, 10, 100, 50, GREEN, fill=True)
    display.rect(10, 70, 100, 50, BLUE, fill=True)
    display.rect(120, 70, 100, 50, YELLOW, fill=True)
    display.line(10, 130, 220, 130, WHITE)
    display.circle(60, 170, 30, MAGENTA, fill=False)
    display.circle(180, 170, 30, CYAN, fill=True)
    display.text("Hello, ILI9341!", 50, 220, WHITE)
    display.text("8080 Parallel Interface", 30, 240, YELLOW)
    display.text("RP2040 PIO Driver", 50, 260, GREEN)
    display.display()
    time.sleep(3)
    
    # Show different rotations
    for r in range(4):
        display.fill(BLACK)
        display.set_rotation(r)
        display.text(f"Rotation {r}", 10, 10, WHITE)
        display.rect(30, 30, WIDTH-60, HEIGHT-60, RED, fill=False)
        display.line(30, 30, WIDTH-30, HEIGHT-30, GREEN)
        display.line(WIDTH-30, 30, 30, HEIGHT-30, BLUE)
        display.circle(WIDTH // 2, HEIGHT // 2, 40, YELLOW, fill=False)
        display.display()
        time.sleep(2)
    
    # Animation demo
    display.set_rotation(0)
    radius = 10
    x, y = WIDTH // 2, HEIGHT // 2
    dx, dy = 2, 3
    colors = [RED, GREEN, BLUE, YELLOW, MAGENTA, CYAN]
    color_idx = 0
    
    for _ in range(300):
        display.fill(BLACK)
        
        # Move the circle
        x += dx
        y += dy
        
        # Bounce off edges
        if x < radius or x > WIDTH - radius:
            dx = -dx
            color_idx = (color_idx + 1) % len(colors)
        if y < radius or y > HEIGHT - radius:
            dy = -dy
            color_idx = (color_idx + 1) % len(colors)
        
        # Draw circle
        display.circle(x, y, radius, colors[color_idx], fill=True)
        
        # Update display
        display.display()
        time.sleep_ms(10)
    
    # Final message
    display.fill(BLACK)
    display.text("Demo complete!", 60, HEIGHT // 2 - 10, WHITE)
    display.display()


if __name__ == "__main__":
    demo()