#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "lcdrgb_rp2.h"

// Include compiled PIO programs
#include "lcdrgb_pio.pio.h"

// RP2040 specific includes
#include "hardware/pio.h"
#include "hardware/dma.h"
#include "hardware/irq.h"
#include "hardware/clocks.h"

// DMA completion handler to trigger frame callback
static void dma_complete_handler(void) {
    // Clear the interrupt request
    dma_hw->ints0 = 1u << LCDRGB_DMA_IRQ0_CHANNEL;
    
    // We would call the Python callback here, but that's not safe from an IRQ
    // Instead, we'll defer to a background task
}

// Initialize PIO and DMA for RGB LCD
void lcdrgb_init_pio(lcdrgb_rgbbus_obj_t *self) {
    // Choose which PIO instance to use (0 or 1)
    self->pio = pio0;
    
    // Select state machine indexes
    self->sm_timing = 0;
    self->sm_data = 1;
    
    // Calculate appropriate clock divider for requested frequency
    uint32_t sys_clk = clock_get_hz(clk_sys);
    float clock_div = (float)sys_clk / (float)self->freq;
    
    // Data format depends on bpp
    uint data_count = (self->bpp == 16) ? 16 : 24;
    if (data_count > self->lane_count) {
        data_count = self->lane_count;
    }
    
    // Find the base pin for data (assuming they're consecutive)
    uint data_base_pin = self->data_pins[0];
    
    // Initialize PIO programs
    lcdrgb_pio_init(self->pio, self->sm_timing, self->sm_data,
                   &self->timing_prog_offset, &self->data_prog_offset,
                   self->hsync_pin, self->vsync_pin, self->de_pin,
                   data_base_pin, data_count);
    
    // Set the clock divider for both state machines
    pio_sm_set_clkdiv(self->pio, self->sm_timing, clock_div);
    pio_sm_set_clkdiv(self->pio, self->sm_data, clock_div);
    
    // Initialize DMA channel for frame data
    self->dma_channel = dma_claim_unused_channel(true);
    self->dma_channel_cfg = dma_claim_unused_channel(true);
    
    // Configure DMA for frame buffer to PIO FIFO transfer
    dma_channel_config c = dma_channel_get_default_config(self->dma_channel);
    channel_config_set_transfer_data_size(&c, (self->bpp == 16) ? DMA_SIZE_16 : DMA_SIZE_32);
    channel_config_set_read_increment(&c, true);
    channel_config_set_write_increment(&c, false);
    channel_config_set_dreq(&c, pio_get_dreq(self->pio, self->sm_data, true));
    
    // We'll set up the actual DMA operation during display updates
}

// Deinitialize PIO and DMA
void lcdrgb_deinit_pio(lcdrgb_rgbbus_obj_t *self) {
    // Stop PIO state machines
    pio_sm_set_enabled(self->pio, self->sm_timing, false);
    pio_sm_set_enabled(self->pio, self->sm_data, false);
    
    // Unclaim DMA channels
    dma_channel_unclaim(self->dma_channel);
    dma_channel_unclaim(self->dma_channel_cfg);
    
    // Clear PIO program memory
    pio_remove_program(self->pio, &rgb_timing_program, self->timing_prog_offset);
    pio_remove_program(self->pio, &rgb_data_program, self->data_prog_offset);
}

// Update display with current framebuffer content
void lcdrgb_update_display(lcdrgb_rgbbus_obj_t *self) {
    if (self->framebuffer == NULL) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("Framebuffer not allocated"));
    }
    
    // Build the timing parameters for the display
    uint32_t hsync_timing = self->hsync_pulse_width;
    uint32_t back_porch = self->hsync_back_porch;
    uint32_t active_width = self->width;
    uint32_t front_porch = self->hsync_front_porch;
    
    // Write timing values to PIO FIFO
    pio_sm_put_blocking(self->pio, self->sm_timing, hsync_timing);
    pio_sm_put_blocking(self->pio, self->sm_timing, back_porch);
    pio_sm_put_blocking(self->pio, self->sm_timing, active_width);
    pio_sm_put_blocking(self->pio, self->sm_timing, front_porch);
    
    // Set up DMA for transferring the framebuffer to PIO
    dma_channel_configure(
        self->dma_channel,
        &dma_channel_get_default_config(self->dma_channel),
        &self->pio->txf[self->sm_data],  // Write to PIO TX FIFO
        self->framebuffer,               // Read from framebuffer
        self->width * self->height,      // Number of pixels
        true                             // Start immediately
    );
    
    // If we have a callback, set up DMA interrupt
    if (self->callback != mp_const_none) {
        dma_channel_set_irq0_enabled(self->dma_channel, true);
        irq_set_exclusive_handler(DMA_IRQ_0, dma_complete_handler);
        irq_set_enabled(DMA_IRQ_0, true);
    }
}