;
; PIO programs for RGB LCD interface
;

; Program for generating HSYNC, VSYNC, and DE signals
.program rgb_timing
.wrap_target
    ; Wait for data from FIFO
    pull block              ; Pull configuration word
    out x, 32               ; Extract first timing parameter (HSYNC)
    
    ; Start HSYNC pulse
    set pins, 0 [31]        ; Assert HSYNC (active low) and delay
    
    ; Wait for back porch
    pull block              ; Pull back porch duration
    out y, 32               ; Extract back porch count
back_porch:
    jmp y--, back_porch [31] ; Delay during back porch
    
    ; Active display period - DE high
    set pins, 2             ; Assert DE, release HSYNC
    pull block              ; Pull active width
    out y, 32               ; Extract active width count
active_line:
    jmp y--, active_line [31] ; Delay during active pixels
    
    ; Front porch
    set pins, 0             ; DE low, HSYNC high
    pull block              ; Pull front porch duration
    out y, 32               ; Extract front porch count
front_porch:
    jmp y--, front_porch [31] ; Delay during front porch
.wrap

; Program for RGB data output
.program rgb_data
.wrap_target
    ; Wait for sync with timing SM
    wait 1 gpio 2           ; Wait for DE high from timing SM
    
    ; Once DE is high, start outputting data pixels
pixel_loop:
    pull block              ; Pull pixel data from FIFO
    out pins, 16            ; Output to data pins (for RGB565)
    wait 0 gpio 2           ; Wait for DE to go low
.wrap

% c-sdk {
// Helper function to initialize PIO programs
static inline void lcdrgb_pio_init(PIO pio, uint sm_timing, uint sm_data, 
                                   uint* timing_offset, uint* data_offset,
                                   uint hsync_pin, uint vsync_pin, uint de_pin, 
                                   uint data_base_pin, uint data_count) {
    
    // Load timing program
    *timing_offset = pio_add_program(pio, &rgb_timing_program);
    
    // Load data program
    *data_offset = pio_add_program(pio, &rgb_data_program);
    
    // Configure timing state machine
    pio_sm_config c_timing = rgb_timing_program_get_default_config(*timing_offset);
    sm_config_set_out_pins(&c_timing, hsync_pin, 2); // HSYNC, VSYNC
    sm_config_set_set_pins(&c_timing, hsync_pin, 3); // HSYNC, VSYNC, DE
    sm_config_set_fifo_join(&c_timing, PIO_FIFO_JOIN_TX); // Join FIFOs to increase size
    
    // Configure data state machine
    pio_sm_config c_data = rgb_data_program_get_default_config(*data_offset);
    sm_config_set_out_pins(&c_data, data_base_pin, data_count); // RGB data pins
    sm_config_set_in_pins(&c_data, de_pin); // Use DE pin as input trigger
    sm_config_set_fifo_join(&c_data, PIO_FIFO_JOIN_TX); // Join FIFOs to increase size
    
    // Initialize pins
    pio_gpio_init(pio, hsync_pin);
    pio_gpio_init(pio, vsync_pin);
    pio_gpio_init(pio, de_pin);
    for (int i = 0; i < data_count; i++) {
        pio_gpio_init(pio, data_base_pin + i);
    }
    
    // Set pin directions
    pio_sm_set_pindirs_with_mask(pio, sm_timing, 
                                 (1u << hsync_pin) | (1u << vsync_pin) | (1u << de_pin),
                                 (1u << hsync_pin) | (1u << vsync_pin) | (1u << de_pin));
    
    pio_sm_set_pindirs_with_mask(pio, sm_data,
                                 ((1u << data_count) - 1) << data_base_pin,
                                 ((1u << data_count) - 1) << data_base_pin);
    
    // Start state machines
    pio_sm_init(pio, sm_timing, *timing_offset, &c_timing);
    pio_sm_init(pio, sm_data, *data_offset, &c_data);
    pio_sm_set_enabled(pio, sm_timing, true);
    pio_sm_set_enabled(pio, sm_data, true);
}
%}