"""
RGB LCD Display Demo
====================

This script demonstrates the usage of the lcdrgb module
for RGB parallel display interfacing on the RP2040.

Connections:
- HSYNC: GPIO2
- VSYNC: GPIO3
- DE: GPIO4
- PCLK: GPIO5
- Data0-7: GPIO6-13 (for 8-bit mode)
- Data0-15: GPIO6-21 (for 16-bit mode)
"""

import time
import math
import random
from machine import Pin
import lcdrgb

# Display resolution (modify according to your display)
WIDTH = 480
HEIGHT = 272

# Color definitions (RGB565 format)
RED = 0xF800
GREEN = 0x07E0
BLUE = 0x001F
BLACK = 0x0000
WHITE = 0xFFFF
YELLOW = 0xFFE0
CYAN = 0x07FF
PURPLE = 0xF81F

def init_display():
    """Initialize RGB LCD display with RP2040."""
    
    # Initialize RGB Bus interface
    # For 8-bit interface (RGB332)
    rgb_bus = lcdrgb.RGBBus(
        # Control pins
        hsync=2,
        vsync=3,
        de=4,
        pclk=5,
        # Data pins (8-bit mode)
        data0=6, data1=7, data2=8, data3=9,
        data4=10, data5=11, data6=12, data7=13,
        # Optional: For 16-bit mode, use the following additional pins
        # data8=14, data9=15, data10=16, data11=17,
        # data12=18, data13=19, data14=20, data15=21,
        # Timing parameters
        freq=10_000_000,  # Pixel clock frequency in Hz
        hsync_pulse_width=41,
        hsync_back_porch=2,
        hsync_front_porch=2,
        vsync_pulse_width=10,
        vsync_back_porch=2,
        vsync_front_porch=2,
        # Signal polarity
        hsync_idle_low=False,
        vsync_idle_low=False,
        de_idle_high=False,
        pclk_idle_high=False,
        pclk_active_low=False
    )
    
    # Initialize with display parameters
    rgb_bus.init(
        WIDTH,           # Width
        HEIGHT,          # Height
        16,              # Bits per pixel (16 for RGB565)
        WIDTH * HEIGHT * 2,  # Buffer size (2 bytes per pixel)
        False,           # RGB565 byte swap
        8,               # Command bits
        8,               # Parameter bits
    )
    
    return rgb_bus

def clear_screen(display, color):
    """Fill the entire screen with a solid color."""
    buffer_size = WIDTH * HEIGHT * 2  # 2 bytes per pixel for RGB565
    buffer = bytearray(buffer_size)
    
    # Fill buffer with color (RGB565 format)
    for i in range(0, buffer_size, 2):
        buffer[i] = color & 0xFF  # Low byte
        buffer[i + 1] = (color >> 8) & 0xFF  # High byte
    
    # Send to display
    display.tx_color(0x2C, buffer, 0, 0, WIDTH-1, HEIGHT-1, 0, True)

def draw_rectangle(display, x, y, w, h, color):
    """Draw a filled rectangle."""
    buffer_size = w * h * 2  # 2 bytes per pixel for RGB565
    buffer = bytearray(buffer_size)
    
    # Fill buffer with color (RGB565 format)
    for i in range(0, buffer_size, 2):
        buffer[i] = color & 0xFF  # Low byte
        buffer[i + 1] = (color >> 8) & 0xFF  # High byte
    
    # Send to display
    display.tx_color(0x2C, buffer, x, y, x+w-1, y+h-1, 0, True)

def draw_color_bars(display):
    """Draw color bars test pattern."""
    bar_width = WIDTH // 8
    colors = [RED, GREEN, BLUE, YELLOW, CYAN, PURPLE, WHITE, BLACK]
    
    for i, color in enumerate(colors):
        x = i * bar_width
        draw_rectangle(display, x, 0, bar_width, HEIGHT, color)

def draw_gradient(display):
    """Draw a color gradient across the screen."""
    buffer_size = WIDTH * HEIGHT * 2  # 2 bytes per pixel for RGB565
    buffer = bytearray(buffer_size)
    
    for y in range(HEIGHT):
        for x in range(WIDTH):
            # Create a gradient pattern
            r = int(31 * x / WIDTH)              # 5 bits for red
            g = int(63 * y / HEIGHT)             # 6 bits for green 
            b = int(31 * (1 - x / WIDTH))        # 5 bits for blue
            
            # Combine into RGB565 format
            color = (r << 11) | (g << 5) | b
            
            # Set pixel in buffer
            idx = (y * WIDTH + x) * 2
            buffer[idx] = color & 0xFF           # Low byte
            buffer[idx + 1] = (color >> 8) & 0xFF  # High byte
    
    # Send to display
    display.tx_color(0x2C, buffer, 0, 0, WIDTH-1, HEIGHT-1, 0, True)

def draw_circles(display):
    """Draw concentric circles with different colors."""
    clear_screen(display, BLACK)
    center_x = WIDTH // 2
    center_y = HEIGHT // 2
    max_radius = min(WIDTH, HEIGHT) // 2 - 10
    colors = [RED, GREEN, BLUE, YELLOW, CYAN, PURPLE]
    
    for r in range(max_radius, 0, -max_radius//len(colors)):
        color = colors[(max_radius - r) % len(colors)]
        
        # Create a simple circle by drawing horizontal lines
        for y in range(HEIGHT):
            for x in range(WIDTH):
                # Check if point is on the circle (approximately)
                dist = math.sqrt((x - center_x)**2 + (y - center_y)**2)
                thickness = 5
                if r - thickness/2 <= dist <= r + thickness/2:
                    # Draw pixel
                    pixel_buffer = bytearray([color & 0xFF, (color >> 8) & 0xFF])
                    display.tx_color(0x2C, pixel_buffer, x, y, x, y, 0, False)
        
        # Update display after each circle
        display.tx_param(0x00)  # Dummy command to trigger update

def animation_demo(display):
    """Show a simple animation of a moving rectangle."""
    rect_width = 80
    rect_height = 60
    rect_color = RED
    bg_color = BLACK
    
    x, y = 0, 0
    dx, dy = 4, 3
    
    for _ in range(100):  # Run animation for 100 frames
        # Clear previous rectangle by drawing background
        draw_rectangle(display, x, y, rect_width, rect_height, bg_color)
        
        # Update position
        x += dx
        y += dy
        
        # Bounce off edges
        if x <= 0 or x + rect_width >= WIDTH:
            dx = -dx
            rect_color = random.choice([RED, GREEN, BLUE, YELLOW, CYAN, PURPLE])
        if y <= 0 or y + rect_height >= HEIGHT:
            dy = -dy
            rect_color = random.choice([RED, GREEN, BLUE, YELLOW, CYAN, PURPLE])
        
        # Draw new rectangle
        draw_rectangle(display, x, y, rect_width, rect_height, rect_color)
        
        time.sleep(0.03)  # Control animation speed

def main():
    # Initialize display
    print("Initializing RGB LCD display...")
    display = init_display()
    
    try:
        # Run demo sequence
        print("Starting demo sequence")
        
        print("1. Clear screen to black")
        clear_screen(display, BLACK)
        time.sleep(1)
        
        print("2. Clear screen to white")
        clear_screen(display, WHITE)
        time.sleep(1)
        
        print("3. Drawing color bars")
        draw_color_bars(display)
        time.sleep(2)
        
        print("4. Drawing gradient pattern")
        draw_gradient(display)
        time.sleep(2)
        
        print("5. Drawing circles")
        draw_circles(display)
        time.sleep(2)
        
        print("6. Running animation")
        animation_demo(display)
        
        print("Demo completed!")
        
    finally:
        # Ensure we clean up properly
        print("Deinitializing display...")
        display.deinit()

if __name__ == "__main__":
    main()