#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "lcdspi.h"

// RP2040 specific includes
#include "hardware/spi.h"
#include "hardware/gpio.h"
#include "hardware/clocks.h"

// SPI handles array (one for each SPI peripheral)
static spi_inst_t* spi_instances[2] = {spi0, spi1};

// Initialize SPI and GPIO pins
static void lcdspi_init_spi(lcdspi_spibus_obj_t *self) {
    // Store SPI instance pointer
    self->spi_handle = spi_instances[self->spi_instance];
    spi_inst_t *spi = (spi_inst_t*)self->spi_handle;
    
    // Initialize GPIO pins
    // DC pin
    gpio_init(self->dc_pin);
    gpio_set_dir(self->dc_pin, GPIO_OUT);
    
    // CS pin if used
    if (self->cs_pin >= 0) {
        gpio_init(self->cs_pin);
        gpio_set_dir(self->cs_pin, GPIO_OUT);
        
        // Initial CS state (inactive)
        gpio_put(self->cs_pin, self->cs_high_active ? 0 : 1);
    }

    // Initialize SPI
    spi_init(spi, self->freq);
    
    // Configure SPI parameters
    // Set bit order
    spi_set_format(spi, 
                  8,                      // 8 bits per transfer
                  self->spi_mode & 0x2 ? 1 : 0,  // CPOL
                  self->spi_mode & 0x1 ? 1 : 0,  // CPHA
                  self->lsb_first ? SPI_LSB_FIRST : SPI_MSB_FIRST);
    
    // The RP2040 SPI pins need to be configured
    if (self->spi_instance == 0) {
        // SPI0 default pins: SCK=18, TX(MOSI)=19, RX(MISO)=16
        gpio_set_function(18, GPIO_FUNC_SPI);  // SCK
        gpio_set_function(19, GPIO_FUNC_SPI);  // MOSI
        gpio_set_function(16, GPIO_FUNC_SPI);  // MISO
    } else {
        // SPI1 default pins: SCK=10, TX(MOSI)=11, RX(MISO)=12
        gpio_set_function(10, GPIO_FUNC_SPI);  // SCK
        gpio_set_function(11, GPIO_FUNC_SPI);  // MOSI
        gpio_set_function(12, GPIO_FUNC_SPI);  // MISO
    }
}

// Deinitialize SPI and GPIO
static void lcdspi_deinit_spi(lcdspi_spibus_obj_t *self) {
    spi_inst_t *spi = (spi_inst_t*)self->spi_handle;
    
    // Deinitialize SPI
    spi_deinit(spi);
    
    // Reset GPIO pins to default state
    if (self->dc_pin >= 0) {
        gpio_set_function(self->dc_pin, GPIO_FUNC_NULL);
    }
    
    if (self->cs_pin >= 0) {
        gpio_set_function(self->cs_pin, GPIO_FUNC_NULL);
    }
    
    // Reset SPI pins to default state
    if (self->spi_instance == 0) {
        gpio_set_function(18, GPIO_FUNC_NULL);  // SCK
        gpio_set_function(19, GPIO_FUNC_NULL);  // MOSI
        gpio_set_function(16, GPIO_FUNC_NULL);  // MISO
    } else {
        gpio_set_function(10, GPIO_FUNC_NULL);  // SCK
        gpio_set_function(11, GPIO_FUNC_NULL);  // MOSI
        gpio_set_function(12, GPIO_FUNC_NULL);  // MISO
    }
}

// Select/deselect chip using CS pin
static void lcdspi_cs_select(lcdspi_spibus_obj_t *self, bool select) {
    if (self->cs_pin >= 0) {
        bool pin_value;
        
        if (self->cs_high_active) {
            pin_value = select ? 1 : 0;
        } else {
            pin_value = select ? 0 : 1;
        }
        
        gpio_put(self->cs_pin, pin_value);
    }
}

// Set DC pin for command or data mode
static void lcdspi_dc_select(lcdspi_spibus_obj_t *self, bool cmd) {
    bool pin_value;
    
    if (self->dc_low_on_data) {
        pin_value = cmd ? 1 : 0;
    } else {
        pin_value = cmd ? 0 : 1;
    }
    
    gpio_put(self->dc_pin, pin_value);
}

// Send data over SPI
static void lcdspi_send_data(lcdspi_spibus_obj_t *self, const uint8_t *data, size_t len) {
    spi_inst_t *spi = (spi_inst_t*)self->spi_handle;
    
    // Split into chunks if needed (for very large transfers)
    const size_t chunk_size = 1024;
    
    for (size_t i = 0; i < len; i += chunk_size) {
        size_t send_len = (i + chunk_size > len) ? (len - i) : chunk_size;
        
        size_t bytes_written = spi_write_blocking(spi, data + i, send_len);
        if (bytes_written != send_len) {
            mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("SPI transmission failed"));
        }
    }
}

// Receive data over SPI
static void lcdspi_receive_data(lcdspi_spibus_obj_t *self, uint8_t *data, size_t len) {
    spi_inst_t *spi = (spi_inst_t*)self->spi_handle;
    
    // Split into chunks if needed
    const size_t chunk_size = 1024;
    
    for (size_t i = 0; i < len; i += chunk_size) {
        size_t recv_len = (i + chunk_size > len) ? (len - i) : chunk_size;
        
        size_t bytes_read = spi_read_blocking(spi, 0xFF, data + i, recv_len);  // Send dummy byte 0xFF
        if (bytes_read != recv_len) {
            mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("SPI reception failed"));
        }
    }
}