#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "py/mperrno.h"
#include "py/objarray.h"
#include "extmod/machine_spi.h"

#include "lcdspi.h"

// RP2040 specific includes
#include "hardware/spi.h"
#include "hardware/gpio.h"

// SPIBus object definition
typedef struct _lcdspi_spibus_obj_t lcdspi_spibus_obj_t;

// Helper function declarations
static void lcdspi_init_spi(lcdspi_spibus_obj_t *self);
static void lcdspi_deinit_spi(lcdspi_spibus_obj_t *self);
static void lcdspi_cs_select(lcdspi_spibus_obj_t *self, bool select);
static void lcdspi_dc_select(lcdspi_spibus_obj_t *self, bool cmd);
static void lcdspi_send_data(lcdspi_spibus_obj_t *self, const uint8_t *data, size_t len);
static void lcdspi_receive_data(lcdspi_spibus_obj_t *self, uint8_t *data, size_t len);

// SPIBus constructor
STATIC mp_obj_t lcdspi_spibus_make_new(const mp_obj_type_t *type, size_t n_args, size_t n_kw, const mp_obj_t *all_args) {
    enum { ARG_spi_bus, ARG_dc, ARG_freq, ARG_cs, ARG_dc_low_on_data, ARG_lsb_first, 
           ARG_cs_high_active, ARG_spi_mode, ARG_dual, ARG_quad, ARG_octal };
    
    static const mp_arg_t allowed_args[] = {
        { MP_QSTR_spi_bus, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_OBJ },
        { MP_QSTR_dc, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_freq, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_cs, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_dc_low_on_data, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_lsb_first, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_cs_high_active, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_spi_mode, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 0} },
        { MP_QSTR_dual, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_quad, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_octal, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
    };
    
    mp_arg_val_t args[MP_ARRAY_SIZE(allowed_args)];
    mp_arg_parse_all_kw_array(n_args, n_kw, all_args, MP_ARRAY_SIZE(allowed_args), allowed_args, args);
    
    // Create new object
    lcdspi_spibus_obj_t *self = mp_obj_malloc(lcdspi_spibus_obj_t, &lcdspi_spibus_type);
    
    // Parse machine.SPI.Bus object
    if (!mp_obj_is_int(args[ARG_spi_bus].u_obj)) {
        mp_raise_ValueError(MP_ERROR_TEXT("spi_bus must be an integer"));
    }
    int spi_id = mp_obj_get_int(args[ARG_spi_bus].u_obj);
    if (spi_id != 0 && spi_id != 1) {
        mp_raise_ValueError(MP_ERROR_TEXT("RP2040 only supports SPI0 (0) or SPI1 (1)"));
    }
    self->spi_instance = spi_id;
    
    // Set GPIO pins and configuration
    self->dc_pin = args[ARG_dc].u_int;
    self->cs_pin = args[ARG_cs].u_int;
    self->freq = args[ARG_freq].u_int;
    self->dc_low_on_data = args[ARG_dc_low_on_data].u_bool;
    self->lsb_first = args[ARG_lsb_first].u_bool;
    self->cs_high_active = args[ARG_cs_high_active].u_bool;
    self->spi_mode = args[ARG_spi_mode].u_int;
    self->dual = args[ARG_dual].u_bool;
    self->quad = args[ARG_quad].u_bool;
    self->octal = args[ARG_octal].u_bool;
    
    // Initialize SPI and GPIO
    lcdspi_init_spi(self);
    
    return MP_OBJ_FROM_PTR(self);
}

// init method
STATIC mp_obj_t lcdspi_spibus_init(size_t n_args, const mp_obj_t *args) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    
    // Parse parameters
    self->width = mp_obj_get_int(args[1]);
    self->height = mp_obj_get_int(args[2]);
    self->bpp = mp_obj_get_int(args[3]);
    self->buffer_size = mp_obj_get_int(args[4]);
    self->rgb565_byte_swap = mp_obj_is_true(args[5]);
    self->cmd_bits = mp_obj_get_int(args[6]);
    self->param_bits = mp_obj_get_int(args[7]);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdspi_spibus_init_obj, 8, 8, lcdspi_spibus_init);

// deinit method
STATIC mp_obj_t lcdspi_spibus_deinit(mp_obj_t self_in) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Deinitialize SPI and GPIO
    lcdspi_deinit_spi(self);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_1(lcdspi_spibus_deinit_obj, lcdspi_spibus_deinit);

// register_callback method
STATIC mp_obj_t lcdspi_spibus_register_callback(mp_obj_t self_in, mp_obj_t callback_in) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Store callback function
    self->callback = MP_OBJ_TO_PTR(callback_in);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_2(lcdspi_spibus_register_callback_obj, lcdspi_spibus_register_callback);

// tx_param method
STATIC mp_obj_t lcdspi_spibus_tx_param(size_t n_args, const mp_obj_t *args) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    
    // Select chip
    lcdspi_cs_select(self, true);
    
    // Send command
    lcdspi_dc_select(self, true);  // Command mode
    lcdspi_send_data(self, &cmd, 1);
    
    // Send parameters if provided
    if (n_args > 2 && args[2] != mp_const_none) {
        mp_buffer_info_t bufinfo;
        mp_get_buffer_raise(args[2], &bufinfo, MP_BUFFER_READ);
        
        lcdspi_dc_select(self, false);  // Data mode
        lcdspi_send_data(self, bufinfo.buf, bufinfo.len);
    }
    
    // Deselect chip
    lcdspi_cs_select(self, false);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdspi_spibus_tx_param_obj, 2, 3, lcdspi_spibus_tx_param);

// tx_color method
STATIC mp_obj_t lcdspi_spibus_tx_color(size_t n_args, const mp_obj_t *args) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(args[2], &bufinfo, MP_BUFFER_READ);
    
    int start_x = mp_obj_get_int(args[3]);
    int start_y = mp_obj_get_int(args[4]);
    int end_x = mp_obj_get_int(args[5]);
    int end_y = mp_obj_get_int(args[6]);
    int rotation = mp_obj_get_int(args[7]);
    bool last_update = mp_obj_is_true(args[8]);
    
    // Select chip
    lcdspi_cs_select(self, true);
    
    // Send command
    lcdspi_dc_select(self, true);  // Command mode
    lcdspi_send_data(self, &cmd, 1);
    
    // Send color data
    lcdspi_dc_select(self, false);  // Data mode
    lcdspi_send_data(self, bufinfo.buf, bufinfo.len);
    
    // Deselect chip
    lcdspi_cs_select(self, false);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdspi_spibus_tx_color_obj, 9, 9, lcdspi_spibus_tx_color);

// rx_param method
STATIC mp_obj_t lcdspi_spibus_rx_param(mp_obj_t self_in, mp_obj_t cmd_in, mp_obj_t params_in) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    uint8_t cmd = mp_obj_get_int(cmd_in);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(params_in, &bufinfo, MP_BUFFER_WRITE);
    
    // Select chip
    lcdspi_cs_select(self, true);
    
    // Send command
    lcdspi_dc_select(self, true);  // Command mode
    lcdspi_send_data(self, &cmd, 1);
    
    // Receive parameters
    lcdspi_dc_select(self, false);  // Data mode
    lcdspi_receive_data(self, bufinfo.buf, bufinfo.len);
    
    // Deselect chip
    lcdspi_cs_select(self, false);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_3(lcdspi_spibus_rx_param_obj, lcdspi_spibus_rx_param);

// get_lane_count method
STATIC mp_obj_t lcdspi_spibus_get_lane_count(mp_obj_t self_in) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Calculate lane count based on mode
    uint8_t lane_count = 1;  // Default is single lane
    
    if (self->octal) {
        lane_count = 8;
    } else if (self->quad) {
        lane_count = 4;
    } else if (self->dual) {
        lane_count = 2;
    }
    
    return MP_OBJ_NEW_SMALL_INT(lane_count);
}
STATIC MP_DEFINE_CONST_FUN_OBJ_1(lcdspi_spibus_get_lane_count_obj, lcdspi_spibus_get_lane_count);

// allocate_framebuffer method
STATIC mp_obj_t lcdspi_spibus_allocate_framebuffer(mp_obj_t self_in, mp_obj_t size_in, mp_obj_t caps_in) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    size_t size = mp_obj_get_int(size_in);
    int caps = mp_obj_get_int(caps_in);
    
    // Free existing framebuffer if there is one
    if (self->framebuffer != NULL) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    // Allocate new framebuffer
    void *buffer = m_malloc(size);
    if (buffer == NULL) {
        return mp_const_none;  // Return None if allocation failed
    }
    
    self->framebuffer = buffer;
    self->framebuffer_size = size;
    
    // Return framebuffer as memoryview
    mp_obj_array_t *mv = m_new_obj(mp_obj_array_t);
    mp_obj_memoryview_init(mv, 'B', 0, size, buffer);
    
    return MP_OBJ_FROM_PTR(mv);
}
STATIC MP_DEFINE_CONST_FUN_OBJ_3(lcdspi_spibus_allocate_framebuffer_obj, lcdspi_spibus_allocate_framebuffer);

// free_framebuffer method
STATIC mp_obj_t lcdspi_spibus_free_framebuffer(mp_obj_t self_in, mp_obj_t framebuffer_in) {
    lcdspi_spibus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Free the framebuffer
    if (self->framebuffer != NULL) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_2(lcdspi_spibus_free_framebuffer_obj, lcdspi_spibus_free_framebuffer);

// SPIBus methods table
STATIC const mp_rom_map_elem_t lcdspi_spibus_locals_dict_table[] = {
    { MP_ROM_QSTR(MP_QSTR_init), MP_ROM_PTR(&lcdspi_spibus_init_obj) },
    { MP_ROM_QSTR(MP_QSTR_deinit), MP_ROM_PTR(&lcdspi_spibus_deinit_obj) },
    { MP_ROM_QSTR(MP_QSTR_register_callback), MP_ROM_PTR(&lcdspi_spibus_register_callback_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_param), MP_ROM_PTR(&lcdspi_spibus_tx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_color), MP_ROM_PTR(&lcdspi_spibus_tx_color_obj) },
    { MP_ROM_QSTR(MP_QSTR_rx_param), MP_ROM_PTR(&lcdspi_spibus_rx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_get_lane_count), MP_ROM_PTR(&lcdspi_spibus_get_lane_count_obj) },
    { MP_ROM_QSTR(MP_QSTR_allocate_framebuffer), MP_ROM_PTR(&lcdspi_spibus_allocate_framebuffer_obj) },
    { MP_ROM_QSTR(MP_QSTR_free_framebuffer), MP_ROM_PTR(&lcdspi_spibus_free_framebuffer_obj) },
};
STATIC MP_DEFINE_CONST_DICT(lcdspi_spibus_locals_dict, lcdspi_spibus_locals_dict_table);

// SPIBus type
const mp_obj_type_t lcdspi_spibus_type = {
    { &mp_type_type },
    .name = MP_QSTR_SPIBus,
    .make_new = lcdspi_spibus_make_new,
    .locals_dict = (mp_obj_dict_t*)&lcdspi_spibus_locals_dict,
};

// Module globals
STATIC const mp_rom_map_elem_t lcdspi_module_globals_table[] = {
    { MP_ROM_QSTR(MP_QSTR___name__), MP_ROM_QSTR(MP_QSTR_lcdspi) },
    { MP_ROM_QSTR(MP_QSTR_SPIBus), MP_ROM_PTR(&lcdspi_spibus_type) },
};
STATIC MP_DEFINE_CONST_DICT(lcdspi_module_globals, lcdspi_module_globals_table);

// Module definition
const mp_obj_module_t lcdspi_user_cmodule = {
    .base = { &mp_type_module },
    .globals = (mp_obj_dict_t*)&lcdspi_module_globals,
};

// Register the module to make it available in Python
MP_REGISTER_MODULE(MP_QSTR_lcdspi, lcdspi_user_cmodule);