import lcdspi
import machine
import time

# Create an SPI bus for a display (using SPI0 on RP2040)
# Note: RP2040 default pins for SPI0 are SCK=18, TX/MOSI=19, RX/MISO=16
display = lcdspi.SPIBus(
    spi_bus=0,          # Use SPI0
    dc=20,              # Data/Command pin
    freq=32000000,      # 32 MHz frequency
    cs=17,              # Chip select pin
    dc_low_on_data=False, # DC high for data
    spi_mode=0          # SPI mode 0 (CPOL=0, CPHA=0)
)

# Initialize display parameters (ILI9341 typical params)
display.init(240, 320, 16, 240*320*2, True, 8, 8)

# Allocate framebuffer for display updates
framebuffer = display.allocate_framebuffer(240*320*2, 0)

# Send initialization commands for ILI9341
# Power control
display.tx_param(0xCB, bytearray([0x39, 0x2C, 0x00, 0x34, 0x02]))
# Power control
display.tx_param(0xCF, bytearray([0x00, 0xC1, 0x30]))
# Driver timing control
display.tx_param(0xE8, bytearray([0x85, 0x00, 0x78]))
# Driver timing control
display.tx_param(0xEA, bytearray([0x00, 0x00]))
# Power control
display.tx_param(0xED, bytearray([0x64, 0x03, 0x12, 0x81]))
# Pump ratio control
display.tx_param(0xF7, bytearray([0x20]))
# Power control
display.tx_param(0xC0, bytearray([0x23]))
# Power control
display.tx_param(0xC1, bytearray([0x10]))
# VCOM control
display.tx_param(0xC5, bytearray([0x3E, 0x28]))
# VCOM control
display.tx_param(0xC7, bytearray([0x86]))

# Memory Access Control
display.tx_param(0x36, bytearray([0x48]))  # Row/col addressing order, RGB pixel order
# Pixel Format Set
display.tx_param(0x3A, bytearray([0x55]))  # 16-bit color format

# Frame Rate Control
display.tx_param(0xB1, bytearray([0x00, 0x18]))

# Display Function Control
display.tx_param(0xB6, bytearray([0x08, 0x82, 0x27]))

# Exit Sleep
display.tx_param(0x11, None)
time.sleep_ms(120)  # Sleep mode exit delay

# Display On
display.tx_param(0x29, None)
time.sleep_ms(10)  # Small delay after turning on display

# Define a function to convert RGB values to 16-bit RGB565 format with byte swap
def rgb565(r, g, b):
    color = ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3)
    # Swap bytes for RGB565_BYTE_SWAP = True
    return ((color >> 8) | ((color & 0xFF) << 8))

# Fill the framebuffer with a test pattern
def fill_test_pattern():
    print("Filling test pattern...")
    view = memoryview(framebuffer)
    width = 240
    height = 320
    
    # Calculate indexes directly into the byte buffer
    for y in range(height):
        for x in range(width):
            index = (y * width + x) * 2
            
            # Create a color gradient
            if y < height // 3:
                # Red gradient
                color = rgb565(255, x * 255 // width, y * 255 // height)
            elif y < 2 * height // 3:
                # Green gradient
                color = rgb565(x * 255 // width, 255, y * 255 // height)
            else:
                # Blue gradient
                color = rgb565(x * 255 // width, y * 255 // height, 255)
                
            view[index] = color & 0xFF
            view[index + 1] = (color >> 8) & 0xFF

# Fill the framebuffer with a solid color
def fill_solid_color(r, g, b):
    print(f"Filling solid color RGB({r},{g},{b})...")
    view = memoryview(framebuffer)
    color = rgb565(r, g, b)
    
    # Calculate the bytes to write
    low_byte = color & 0xFF
    high_byte = (color >> 8) & 0xFF
    
    # Fill the entire buffer with the color
    for i in range(0, len(framebuffer), 2):
        view[i] = low_byte
        view[i + 1] = high_byte

# Set column and row address windows
def set_window(x_start, y_start, x_end, y_end):
    # Set column address
    display.tx_param(0x2A, bytearray([
        (x_start >> 8) & 0xFF, x_start & 0xFF,
        (x_end >> 8) & 0xFF, x_end & 0xFF
    ]))
    
    # Set row address
    display.tx_param(0x2B, bytearray([
        (y_start >> 8) & 0xFF, y_start & 0xFF,
        (y_end >> 8) & 0xFF, y_end & 0xFF
    ]))

# Test the display with different patterns
print("Testing ILI9341 display...")

# Display red screen
fill_solid_color(255, 0, 0)
set_window(0, 0, 239, 319)
display.tx_color(0x2C, framebuffer, 0, 0, 239, 319, 0, True)
print("Red screen displayed")
time.sleep(1)

# Display green screen
fill_solid_color(0, 255, 0)
set_window(0, 0, 239, 319)
display.tx_color(0x2C, framebuffer, 0, 0, 239, 319, 0, True)
print("Green screen displayed")
time.sleep(1)

# Display blue screen
fill_solid_color(0, 0, 255)
set_window(0, 0, 239, 319)
display.tx_color(0x2C, framebuffer, 0, 0, 239, 319, 0, True)
print("Blue screen displayed")
time.sleep(1)

# Display color gradient test pattern
fill_test_pattern()
set_window(0, 0, 239, 319)
display.tx_color(0x2C, framebuffer, 0, 0, 239, 319, 0, True)
print("Color test pattern displayed")

print("Test complete")