#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "lcdi2c.h"

// STM32 specific includes
#include "stm32f4xx_hal.h"

// I2C handles array (one for each I2C peripheral)
static I2C_HandleTypeDef i2c_handles[3] = {0};

// Initialize I2C and GPIO pins
static void lcdi2c_init_i2c(lcdi2c_i2cbus_obj_t *self) {
    // Configure GPIO pins for SDA and SCL
    GPIO_InitTypeDef GPIO_InitStruct = {0};
    
    // Enable GPIO clocks (this assumes the pins are on GPIOB, adjust as needed)
    __HAL_RCC_GPIOB_CLK_ENABLE();
    
    // Configure SDA pin
    GPIO_InitStruct.Pin = (1 << self->sda_pin);
    GPIO_InitStruct.Mode = GPIO_MODE_AF_OD;  // Alternate function, open-drain
    GPIO_InitStruct.Pull = self->sda_pullup ? GPIO_PULLUP : GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_HIGH;
    GPIO_InitStruct.Alternate = GPIO_AF4_I2C1;  // Set appropriate AF based on I2C instance
    HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);
    
    // Configure SCL pin
    GPIO_InitStruct.Pin = (1 << self->scl_pin);
    GPIO_InitStruct.Pull = self->scl_pullup ? GPIO_PULLUP : GPIO_NOPULL;
    HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);
    
    // Configure and initialize I2C
    I2C_HandleTypeDef *hi2c = &i2c_handles[self->host];
    self->i2c_handle = hi2c;
    
    // Enable I2C clock based on the host number
    switch (self->host) {
        case 0:
            __HAL_RCC_I2C1_CLK_ENABLE();
            hi2c->Instance = I2C1;
            break;
        case 1:
            __HAL_RCC_I2C2_CLK_ENABLE();
            hi2c->Instance = I2C2;
            break;
        case 2:
            __HAL_RCC_I2C3_CLK_ENABLE();
            hi2c->Instance = I2C3;
            break;
        default:
            mp_raise_ValueError(MP_ERROR_TEXT("Invalid I2C host"));
    }
    
    // Configure I2C parameters
    hi2c->Init.ClockSpeed = self->freq;
    hi2c->Init.DutyCycle = I2C_DUTYCYCLE_2;
    hi2c->Init.OwnAddress1 = 0;
    hi2c->Init.AddressingMode = I2C_ADDRESSINGMODE_7BIT;
    hi2c->Init.DualAddressMode = I2C_DUALADDRESS_DISABLE;
    hi2c->Init.OwnAddress2 = 0;
    hi2c->Init.GeneralCallMode = I2C_GENERALCALL_DISABLE;
    hi2c->Init.NoStretchMode = I2C_NOSTRETCH_DISABLE;
    
    // Initialize I2C
    if (HAL_I2C_Init(hi2c) != HAL_OK) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("Failed to initialize I2C"));
    }
}

// Deinitialize I2C and GPIO
static void lcdi2c_deinit_i2c(lcdi2c_i2cbus_obj_t *self) {
    I2C_HandleTypeDef *hi2c = (I2C_HandleTypeDef *)self->i2c_handle;
    
    // Deinitialize I2C
    if (hi2c != NULL) {
        HAL_I2C_DeInit(hi2c);
    }
    
    // Deinitialize GPIO pins
    HAL_GPIO_DeInit(GPIOB, (1 << self->sda_pin));
    HAL_GPIO_DeInit(GPIOB, (1 << self->scl_pin));
}

// Prepare control byte for I2C transaction
static uint8_t lcdi2c_prepare_control_byte(lcdi2c_i2cbus_obj_t *self, bool is_command) {
    if (self->disable_control_phase) {
        return 0; // No control byte used
    }
    
    uint8_t control_byte = 0;
    
    // Set or clear the DC bit based on whether this is a command or data
    bool dc_bit_value;
    if (self->dc_low_on_data) {
        dc_bit_value = is_command ? 1 : 0;
    } else {
        dc_bit_value = is_command ? 0 : 1;
    }
    
    if (dc_bit_value) {
        control_byte |= (1 << self->dc_bit_offset);
    }
    
    return control_byte;
}

// Send data over I2C with control byte(s)
static void lcdi2c_send_data(lcdi2c_i2cbus_obj_t *self, const uint8_t *data, size_t len, bool is_command) {
    I2C_HandleTypeDef *hi2c = (I2C_HandleTypeDef *)self->i2c_handle;
    
    if (self->disable_control_phase) {
        // Send data directly without control byte
        if (HAL_I2C_Master_Transmit(hi2c, self->addr << 1, (uint8_t *)data, len, HAL_MAX_DELAY) != HAL_OK) {
            mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("I2C transmission failed"));
        }
        return;
    }
    
    // Prepare control byte
    uint8_t control_byte = lcdi2c_prepare_control_byte(self, is_command);
    
    // Allocate buffer for control byte(s) + data
    uint8_t *buffer = m_new(uint8_t, self->control_phase_bytes + len);
    if (buffer == NULL) {
        mp_raise_msg(&mp_type_MemoryError, MP_ERROR_TEXT("Failed to allocate I2C buffer"));
    }
    
    // Fill control bytes
    for (int i = 0; i < self->control_phase_bytes; i++) {
        buffer[i] = control_byte;
    }
    
    // Copy data
    memcpy(buffer + self->control_phase_bytes, data, len);
    
    // Send the complete buffer
    if (HAL_I2C_Master_Transmit(hi2c, self->addr << 1, buffer, self->control_phase_bytes + len, HAL_MAX_DELAY) != HAL_OK) {
        m_del(uint8_t, buffer, self->control_phase_bytes + len);
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("I2C transmission failed"));
    }
    
    // Free the buffer
    m_del(uint8_t, buffer, self->control_phase_bytes + len);
}

// Receive data over I2C
static void lcdi2c_receive_data(lcdi2c_i2cbus_obj_t *self, uint8_t *data, size_t len, bool is_command) {
    I2C_HandleTypeDef *hi2c = (I2C_HandleTypeDef *)self->i2c_handle;
    
    if (!self->disable_control_phase) {
        // Send control byte first
        uint8_t control_byte = lcdi2c_prepare_control_byte(self, is_command);
        
        if (HAL_I2C_Master_Transmit(hi2c, self->addr << 1, &control_byte, 1, HAL_MAX_DELAY) != HAL_OK) {
            mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("I2C control byte transmission failed"));
        }
    }
    
    // Now receive the data
    if (HAL_I2C_Master_Receive(hi2c, self->addr << 1, data, len, HAL_MAX_DELAY) != HAL_OK) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("I2C reception failed"));
    }
}