#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "py/mperrno.h"
#include "py/objarray.h"

#include "lcdi2c.h"

// STM32 specific includes
#include "stm32f4xx_hal.h"

// Forward declarations for helper functions
static void lcdi2c_init_i2c(lcdi2c_i2cbus_obj_t *self);
static void lcdi2c_deinit_i2c(lcdi2c_i2cbus_obj_t *self);
static void lcdi2c_send_data(lcdi2c_i2cbus_obj_t *self, const uint8_t *data, size_t len, bool is_command);
static void lcdi2c_receive_data(lcdi2c_i2cbus_obj_t *self, uint8_t *data, size_t len, bool is_command);

// I2CBus constructor
STATIC mp_obj_t lcdi2c_i2cbus_make_new(const mp_obj_type_t *type, size_t n_args, size_t n_kw, const mp_obj_t *all_args) {
    enum { ARG_sda, ARG_scl, ARG_addr, ARG_host, ARG_control_phase_bytes, ARG_dc_bit_offset, 
           ARG_freq, ARG_dc_low_on_data, ARG_sda_pullup, ARG_scl_pullup, ARG_disable_control_phase };
    
    static const mp_arg_t allowed_args[] = {
        { MP_QSTR_sda, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_scl, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_addr, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_host, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 0} },
        { MP_QSTR_control_phase_bytes, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 1} },
        { MP_QSTR_dc_bit_offset, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 6} },
        { MP_QSTR_freq, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 10000000} },
        { MP_QSTR_dc_low_on_data, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_sda_pullup, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = true} },
        { MP_QSTR_scl_pullup, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = true} },
        { MP_QSTR_disable_control_phase, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
    };
    
    mp_arg_val_t args[MP_ARRAY_SIZE(allowed_args)];
    mp_arg_parse_all_kw_array(n_args, n_kw, all_args, MP_ARRAY_SIZE(allowed_args), allowed_args, args);
    
    // Create new object
    lcdi2c_i2cbus_obj_t *self = mp_obj_malloc(lcdi2c_i2cbus_obj_t, &lcdi2c_i2cbus_type);
    
    // Set I2C configuration
    self->sda_pin = args[ARG_sda].u_int;
    self->scl_pin = args[ARG_scl].u_int;
    self->addr = args[ARG_addr].u_int;
    self->host = args[ARG_host].u_int;
    self->control_phase_bytes = args[ARG_control_phase_bytes].u_int;
    self->dc_bit_offset = args[ARG_dc_bit_offset].u_int;
    self->freq = args[ARG_freq].u_int;
    self->dc_low_on_data = args[ARG_dc_low_on_data].u_bool;
    self->sda_pullup = args[ARG_sda_pullup].u_bool;
    self->scl_pullup = args[ARG_scl_pullup].u_bool;
    self->disable_control_phase = args[ARG_disable_control_phase].u_bool;
    
    // Initialize I2C
    lcdi2c_init_i2c(self);
    
    return MP_OBJ_FROM_PTR(self);
}

// init method
STATIC mp_obj_t lcdi2c_i2cbus_init(size_t n_args, const mp_obj_t *args) {
    lcdi2c_i2cbus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    
    // Parse parameters
    self->width = mp_obj_get_int(args[1]);
    self->height = mp_obj_get_int(args[2]);
    self->bpp = mp_obj_get_int(args[3]);
    self->buffer_size = mp_obj_get_int(args[4]);
    self->rgb565_byte_swap = mp_obj_is_true(args[5]);
    self->cmd_bits = mp_obj_get_int(args[6]);
    self->param_bits = mp_obj_get_int(args[7]);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdi2c_i2cbus_init_obj, 8, 8, lcdi2c_i2cbus_init);

// deinit method
STATIC mp_obj_t lcdi2c_i2cbus_deinit(mp_obj_t self_in) {
    lcdi2c_i2cbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Deinitialize I2C
    lcdi2c_deinit_i2c(self);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_1(lcdi2c_i2cbus_deinit_obj, lcdi2c_i2cbus_deinit);

// register_callback method
STATIC mp_obj_t lcdi2c_i2cbus_register_callback(mp_obj_t self_in, mp_obj_t callback_in) {
    lcdi2c_i2cbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Store callback function
    self->callback = MP_OBJ_TO_PTR(callback_in);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_2(lcdi2c_i2cbus_register_callback_obj, lcdi2c_i2cbus_register_callback);

// tx_param method
STATIC mp_obj_t lcdi2c_i2cbus_tx_param(size_t n_args, const mp_obj_t *args) {
    lcdi2c_i2cbus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    
    // Send command
    lcdi2c_send_data(self, &cmd, 1, true);
    
    // Send parameters if provided
    if (n_args > 2 && args[2] != mp_const_none) {
        mp_buffer_info_t bufinfo;
        mp_get_buffer_raise(args[2], &bufinfo, MP_BUFFER_READ);
        
        lcdi2c_send_data(self, bufinfo.buf, bufinfo.len, false);
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdi2c_i2cbus_tx_param_obj, 2, 3, lcdi2c_i2cbus_tx_param);

// tx_color method
STATIC mp_obj_t lcdi2c_i2cbus_tx_color(size_t n_args, const mp_obj_t *args) {
    lcdi2c_i2cbus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(args[2], &bufinfo, MP_BUFFER_READ);
    
    int start_x = mp_obj_get_int(args[3]);
    int start_y = mp_obj_get_int(args[4]);
    int end_x = mp_obj_get_int(args[5]);
    int end_y = mp_obj_get_int(args[6]);
    int rotation = mp_obj_get_int(args[7]);
    bool last_update = mp_obj_is_true(args[8]);
    
    // Send command
    lcdi2c_send_data(self, &cmd, 1, true);
    
    // Send color data
    lcdi2c_send_data(self, bufinfo.buf, bufinfo.len, false);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdi2c_i2cbus_tx_color_obj, 9, 9, lcdi2c_i2cbus_tx_color);

// rx_param method
STATIC mp_obj_t lcdi2c_i2cbus_rx_param(mp_obj_t self_in, mp_obj_t cmd_in, mp_obj_t params_in) {
    lcdi2c_i2cbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    uint8_t cmd = mp_obj_get_int(cmd_in);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(params_in, &bufinfo, MP_BUFFER_WRITE);
    
    // Send command
    lcdi2c_send_data(self, &cmd, 1, true);
    
    // Receive parameters
    lcdi2c_receive_data(self, bufinfo.buf, bufinfo.len, false);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_3(lcdi2c_i2cbus_rx_param_obj, lcdi2c_i2cbus_rx_param);

// get_lane_count method
STATIC mp_obj_t lcdi2c_i2cbus_get_lane_count(mp_obj_t self_in) {
    // I2C is always single lane
    return MP_OBJ_NEW_SMALL_INT(1);
}
STATIC MP_DEFINE_CONST_FUN_OBJ_1(lcdi2c_i2cbus_get_lane_count_obj, lcdi2c_i2cbus_get_lane_count);

// allocate_framebuffer method
STATIC mp_obj_t lcdi2c_i2cbus_allocate_framebuffer(mp_obj_t self_in, mp_obj_t size_in, mp_obj_t caps_in) {
    lcdi2c_i2cbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    size_t size = mp_obj_get_int(size_in);
    int caps = mp_obj_get_int(caps_in);
    
    // Free existing framebuffer if there is one
    if (self->framebuffer != NULL) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    // Allocate new framebuffer
    void *buffer = m_malloc(size);
    if (buffer == NULL) {
        return mp_const_none;  // Return None if allocation failed
    }
    
    self->framebuffer = buffer;
    self->framebuffer_size = size;
    
    // Return framebuffer as memoryview
    mp_obj_array_t *mv = m_new_obj(mp_obj_array_t);
    mp_obj_memoryview_init(mv, 'B', 0, size, buffer);
    
    return MP_OBJ_FROM_PTR(mv);
}
STATIC MP_DEFINE_CONST_FUN_OBJ_3(lcdi2c_i2cbus_allocate_framebuffer_obj, lcdi2c_i2cbus_allocate_framebuffer);

// free_framebuffer method
STATIC mp_obj_t lcdi2c_i2cbus_free_framebuffer(mp_obj_t self_in, mp_obj_t framebuffer_in) {
    lcdi2c_i2cbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Free the framebuffer
    if (self->framebuffer != NULL) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_2(lcdi2c_i2cbus_free_framebuffer_obj, lcdi2c_i2cbus_free_framebuffer);

// I2CBus methods table
STATIC const mp_rom_map_elem_t lcdi2c_i2cbus_locals_dict_table[] = {
    { MP_ROM_QSTR(MP_QSTR_init), MP_ROM_PTR(&lcdi2c_i2cbus_init_obj) },
    { MP_ROM_QSTR(MP_QSTR_deinit), MP_ROM_PTR(&lcdi2c_i2cbus_deinit_obj) },
    { MP_ROM_QSTR(MP_QSTR_register_callback), MP_ROM_PTR(&lcdi2c_i2cbus_register_callback_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_param), MP_ROM_PTR(&lcdi2c_i2cbus_tx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_color), MP_ROM_PTR(&lcdi2c_i2cbus_tx_color_obj) },
    { MP_ROM_QSTR(MP_QSTR_rx_param), MP_ROM_PTR(&lcdi2c_i2cbus_rx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_get_lane_count), MP_ROM_PTR(&lcdi2c_i2cbus_get_lane_count_obj) },
    { MP_ROM_QSTR(MP_QSTR_allocate_framebuffer), MP_ROM_PTR(&lcdi2c_i2cbus_allocate_framebuffer_obj) },
    { MP_ROM_QSTR(MP_QSTR_free_framebuffer), MP_ROM_PTR(&lcdi2c_i2cbus_free_framebuffer_obj) },
};
STATIC MP_DEFINE_CONST_DICT(lcdi2c_i2cbus_locals_dict, lcdi2c_i2cbus_locals_dict_table);

// I2CBus type
const mp_obj_type_t lcdi2c_i2cbus_type = {
    { &mp_type_type },
    .name = MP_QSTR_I2CBus,
    .make_new = lcdi2c_i2cbus_make_new,
    .locals_dict = (mp_obj_dict_t*)&lcdi2c_i2cbus_locals_dict,
};

// Module globals
STATIC const mp_rom_map_elem_t lcdi2c_module_globals_table[] = {
    { MP_ROM_QSTR(MP_QSTR___name__), MP_ROM_QSTR(MP_QSTR_lcdi2c) },
    { MP_ROM_QSTR(MP_QSTR_I2CBus), MP_ROM_PTR(&lcdi2c_i2cbus_type) },
};
STATIC MP_DEFINE_CONST_DICT(lcdi2c_module_globals, lcdi2c_module_globals_table);

// Module definition
const mp_obj_module_t lcdi2c_user_cmodule = {
    .base = { &mp_type_module },
    .globals = (mp_obj_dict_t*)&lcdi2c_module_globals,
};

// Register the module to make it available in Python
MP_REGISTER_MODULE(MP_QSTR_lcdi2c, lcdi2c_user_cmodule);