#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "lcdi8080.h"

// STM32 specific includes
#include "stm32f4xx_hal.h"

// Helper function declarations
static void lcdi8080_init_fmc(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_deinit_fmc(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_cs_select(lcdi8080_i80bus_obj_t *self, bool select);
static void lcdi8080_dc_set(lcdi8080_i80bus_obj_t *self, bool cmd, bool dummy, bool data);
static void lcdi8080_write_data(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len);
static void lcdi8080_read_data(lcdi8080_i80bus_obj_t *self, uint8_t *data, size_t len);

// Initialize FMC for 8080 parallel interface
static void lcdi8080_init_fmc(lcdi8080_i80bus_obj_t *self) {
    // Enable FMC clock
    __HAL_RCC_FMC_CLK_ENABLE();
    
    // Enable GPIO clocks for all used pins
    __HAL_RCC_GPIOD_CLK_ENABLE();  // Typically FMC pins are on GPIOD and GPIOE
    __HAL_RCC_GPIOE_CLK_ENABLE();
    
    // Configure control pins (DC, WR, CS) as GPIO if not using FMC pins
    GPIO_InitTypeDef GPIO_InitStruct = {0};
    
    // Configure DC pin
    GPIO_InitStruct.Pin = (1 << self->dc_pin);
    GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
    GPIO_InitStruct.Pull = GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
    HAL_GPIO_Init(GPIOD, &GPIO_InitStruct);  // Assuming DC is on GPIOD
    
    // Set initial DC state
    HAL_GPIO_WritePin(GPIOD, (1 << self->dc_pin), 
                      self->dc_idle_high ? GPIO_PIN_SET : GPIO_PIN_RESET);
    
    // Configure FMC pins for data and control
    GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
    GPIO_InitStruct.Pull = GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
    GPIO_InitStruct.Alternate = GPIO_AF12_FMC;
    
    // Configure data pins
    uint16_t pins_d = 0;
    uint16_t pins_e = 0;
    
    // Organize which pins go to which GPIO port
    for (int i = 0; i < self->data_width; i++) {
        if (self->data_pins[i] < 16) { // Assuming pins 0-15 are on GPIOD
            pins_d |= (1 << self->data_pins[i]);
        } else { // Assuming pins 16-31 are on GPIOE
            pins_e |= (1 << (self->data_pins[i] - 16));
        }
    }
    
    // Configure GPIOD pins
    if (pins_d != 0) {
        GPIO_InitStruct.Pin = pins_d;
        HAL_GPIO_Init(GPIOD, &GPIO_InitStruct);
    }
    
    // Configure GPIOE pins
    if (pins_e != 0) {
        GPIO_InitStruct.Pin = pins_e;
        HAL_GPIO_Init(GPIOE, &GPIO_InitStruct);
    }
    
    // Configure FMC NOR/SRAM bank
    FMC_NORSRAM_InitTypeDef FMC_InitStruct = {0};
    FMC_NORSRAM_TimingTypeDef FMC_TimingStruct = {0};
    
    // Calculate timing based on requested frequency
    uint32_t hclk = HAL_RCC_GetHCLKFreq();
    uint32_t period_ns = 1000000000 / self->freq;
    
    // Calculate timing parameters in FMC clock cycles
    uint32_t cycle_ns = 1000000000 / hclk;
    uint8_t addressSetupTime = period_ns / (2 * cycle_ns);
    uint8_t dataSetupTime = period_ns / (2 * cycle_ns);
    
    // Ensure minimum values
    if (addressSetupTime == 0) addressSetupTime = 1;
    if (dataSetupTime == 0) dataSetupTime = 1;
    
    // Configure timing for the FMC NOR/SRAM
    FMC_TimingStruct.AddressSetupTime = addressSetupTime;
    FMC_TimingStruct.AddressHoldTime = 1;
    FMC_TimingStruct.DataSetupTime = dataSetupTime;
    FMC_TimingStruct.BusTurnAroundDuration = 0;
    FMC_TimingStruct.CLKDivision = 2;
    FMC_TimingStruct.DataLatency = 2;
    FMC_TimingStruct.AccessMode = FMC_ACCESS_MODE_A;
    
    // Configure FMC NOR/SRAM init structure
    FMC_InitStruct.NSBank = FMC_NORSRAM_BANK1;
    FMC_InitStruct.DataAddressMux = FMC_DATA_ADDRESS_MUX_DISABLE;
    FMC_InitStruct.MemoryType = FMC_MEMORY_TYPE_SRAM;
    FMC_InitStruct.MemoryDataWidth = (self->data_width <= 8) ? 
                                       FMC_NORSRAM_MEM_BUS_WIDTH_8 : 
                                       FMC_NORSRAM_MEM_BUS_WIDTH_16;
    FMC_InitStruct.BurstAccessMode = FMC_BURST_ACCESS_MODE_DISABLE;
    FMC_InitStruct.WaitSignalPolarity = FMC_WAIT_SIGNAL_POLARITY_LOW;
    FMC_InitStruct.WrapMode = FMC_WRAP_MODE_DISABLE;
    FMC_InitStruct.WaitSignalActive = FMC_WAIT_TIMING_BEFORE_WS;
    FMC_InitStruct.WriteOperation = FMC_WRITE_OPERATION_ENABLE;
    FMC_InitStruct.WaitSignal = FMC_WAIT_SIGNAL_DISABLE;
    FMC_InitStruct.ExtendedMode = FMC_EXTENDED_MODE_DISABLE;
    FMC_InitStruct.AsynchronousWait = FMC_ASYNCHRONOUS_WAIT_DISABLE;
    FMC_InitStruct.WriteBurst = FMC_WRITE_BURST_DISABLE;
    FMC_InitStruct.ContinuousClock = FMC_CONTINUOUS_CLOCK_SYNC_ONLY;
    
    // Initialize FMC controller
    if (HAL_SRAM_Init(&hsram1, &FMC_TimingStruct, NULL) != HAL_OK) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("Failed to initialize FMC"));
    }
    
    // Store FMC bank pointer for future use
    self->fmc_bank = FMC_NORSRAM_DEVICE->BTCR[(FMC_InitStruct.NSBank - 1) * 2];
}

// Deinitialize FMC
static void lcdi8080_deinit_fmc(lcdi8080_i80bus_obj_t *self) {
    // Deinitialize FMC
    HAL_SRAM_DeInit(&hsram1);
    
    // Deinitialize GPIO pins
    uint16_t pins_d = 0;
    uint16_t pins_e = 0;
    
    // Collect pins on each port
    for (int i = 0; i < self->data_width; i++) {
        if (self->data_pins[i] < 16) {
            pins_d |= (1 << self->data_pins[i]);
        } else {
            pins_e |= (1 << (self->data_pins[i] - 16));
        }
    }
    
    // Deinitialize GPIOD pins
    if (pins_d != 0) {
        HAL_GPIO_DeInit(GPIOD, pins_d);
    }
    
    // Deinitialize GPIOE pins
    if (pins_e != 0) {
        HAL_GPIO_DeInit(GPIOE, pins_e);
    }
    
    // Deinitialize control pins
    HAL_GPIO_DeInit(GPIOD, (1 << self->dc_pin));
}

// Set CS pin state
static void lcdi8080_cs_select(lcdi8080_i80bus_obj_t *self, bool select) {
    if (self->cs_pin >= 0) {
        GPIO_PinState state = (select == self->cs_active_high) ? GPIO_PIN_SET : GPIO_PIN_RESET;
        HAL_GPIO_WritePin(GPIOD, (1 << self->cs_pin), state); // Assuming CS is on GPIOD
    }
}

// Set DC pin state based on type (command, dummy, or data)
static void lcdi8080_dc_set(lcdi8080_i80bus_obj_t *self, bool cmd, bool dummy, bool data) {
    GPIO_PinState state;
    
    if (cmd) {
        state = self->dc_cmd_high ? GPIO_PIN_SET : GPIO_PIN_RESET;
    } else if (dummy) {
        state = self->dc_dummy_high ? GPIO_PIN_SET : GPIO_PIN_RESET;
    } else if (data) {
        state = self->dc_data_high ? GPIO_PIN_SET : GPIO_PIN_RESET;
    } else {
        state = self->dc_idle_high ? GPIO_PIN_SET : GPIO_PIN_RESET;
    }
    
    HAL_GPIO_WritePin(GPIOD, (1 << self->dc_pin), state);
}

// Write data to LCD using FMC
static void lcdi8080_write_data(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len) {
    uint32_t address = 0x60000000; // Base address for FMC Bank 1
    
    if (self->data_width <= 8) {
        // 8-bit bus mode
        for (size_t i = 0; i < len; i++) {
            *(__IO uint8_t *)(address) = data[i];
        }
    } else {
        // 16-bit bus mode
        uint16_t *data16 = (uint16_t *)data;
        size_t len16 = (len + 1) / 2; // Round up to handle odd number of bytes
        
        if (self->swap_color_bytes && len >= 2) {
            // With byte swapping
            for (size_t i = 0; i < len16; i++) {
                uint16_t value = data16[i];
                if (self->swap_color_bytes) {
                    value = ((value & 0xFF) << 8) | ((value >> 8) & 0xFF);
                }
                *(__IO uint16_t *)(address) = value;
            }
        } else {
            // Without byte swapping
            for (size_t i = 0; i < len16; i++) {
                *(__IO uint16_t *)(address) = data16[i];
            }
        }
    }
}

// Read data from LCD using FMC
static void lcdi8080_read_data(lcdi8080_i80bus_obj_t *self, uint8_t *data, size_t len) {
    uint32_t address = 0x60000000; // Base address for FMC Bank 1
    
    if (self->data_width <= 8) {
        // 8-bit bus mode
        for (size_t i = 0; i < len; i++) {
            data[i] = *(__IO uint8_t *)(address);
        }
    } else {
        // 16-bit bus mode
        uint16_t *data16 = (uint16_t *)data;
        size_t len16 = (len + 1) / 2; // Round up to handle odd number of bytes
        
        for (size_t i = 0; i < len16; i++) {
            uint16_t value = *(__IO uint16_t *)(address);
            if (self->swap_color_bytes) {
                value = ((value & 0xFF) << 8) | ((value >> 8) & 0xFF);
            }
            data16[i] = value;
        }
    }
}