#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "py/mperrno.h"
#include "py/objarray.h"

#include "lcdi8080.h"

// STM32 specific includes
#include "stm32f4xx_hal.h"

// Required FMC SRAM handle
SRAM_HandleTypeDef hsram1;

// Helper function declarations
static void lcdi8080_init_fmc(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_deinit_fmc(lcdi8080_i80bus_obj_t *self);
static void lcdi8080_cs_select(lcdi8080_i80bus_obj_t *self, bool select);
static void lcdi8080_dc_set(lcdi8080_i80bus_obj_t *self, bool cmd, bool dummy, bool data);
static void lcdi8080_write_data(lcdi8080_i80bus_obj_t *self, const uint8_t *data, size_t len);
static void lcdi8080_read_data(lcdi8080_i80bus_obj_t *self, uint8_t *data, size_t len);

// I80Bus constructor
STATIC mp_obj_t lcdi8080_i80bus_make_new(const mp_obj_type_t *type, size_t n_args, size_t n_kw, const mp_obj_t *all_args) {
    // Define kwargs for I80Bus constructor
    enum { 
        ARG_dc, ARG_wr, ARG_data0, ARG_data1, ARG_data2, ARG_data3, ARG_data4, ARG_data5, 
        ARG_data6, ARG_data7, ARG_data8, ARG_data9, ARG_data10, ARG_data11, ARG_data12, 
        ARG_data13, ARG_data14, ARG_data15, ARG_cs, ARG_freq, ARG_dc_idle_high, ARG_dc_cmd_high,
        ARG_dc_dummy_high, ARG_dc_data_high, ARG_cs_active_high, ARG_reverse_color_bits,
        ARG_swap_color_bytes, ARG_pclk_active_low, ARG_pclk_idle_low
    };
    
    static const mp_arg_t allowed_args[] = {
        { MP_QSTR_dc, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_wr, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data0, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data1, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data2, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data3, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data4, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data5, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data6, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data7, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data8, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data9, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data10, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data11, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data12, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data13, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data14, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data15, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_cs, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_freq, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 10000000} },
        { MP_QSTR_dc_idle_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_cmd_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_dummy_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_dc_data_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = true} },
        { MP_QSTR_cs_active_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_reverse_color_bits, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_swap_color_bytes, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_pclk_active_low, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_pclk_idle_low, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
    };
    
    mp_arg_val_t args[MP_ARRAY_SIZE(allowed_args)];
    mp_arg_parse_all_kw_array(n_args, n_kw, all_args, MP_ARRAY_SIZE(allowed_args), allowed_args, args);
    
    // Create new I80Bus object
    lcdi8080_i80bus_obj_t *self = mp_obj_malloc(lcdi8080_i80bus_obj_t, &lcdi8080_i80bus_type);
    
    // Set pins and configuration
    self->dc_pin = args[ARG_dc].u_int;
    self->wr_pin = args[ARG_wr].u_int;
    self->cs_pin = args[ARG_cs].u_int;
    
    // Set data pins
    self->data_pins[0] = args[ARG_data0].u_int;
    self->data_pins[1] = args[ARG_data1].u_int;
    self->data_pins[2] = args[ARG_data2].u_int;
    self->data_pins[3] = args[ARG_data3].u_int;
    self->data_pins[4] = args[ARG_data4].u_int;
    self->data_pins[5] = args[ARG_data5].u_int;
    self->data_pins[6] = args[ARG_data6].u_int;
    self->data_pins[7] = args[ARG_data7].u_int;
    
    // Determine bus width (8-bit or 16-bit)
    self->data_width = 8;  // Default to 8-bit
    
    // Check if any 16-bit data pins are defined
    if (args[ARG_data8].u_int != -1) {
        self->data_pins[8] = args[ARG_data8].u_int;
        self->data_width = 16;
    }
    if (args[ARG_data9].u_int != -1) {
        self->data_pins[9] = args[ARG_data9].u_int;
    }
    if (args[ARG_data10].u_int != -1) {
        self->data_pins[10] = args[ARG_data10].u_int;
    }
    if (args[ARG_data11].u_int != -1) {
        self->data_pins[11] = args[ARG_data11].u_int;
    }
    if (args[ARG_data12].u_int != -1) {
        self->data_pins[12] = args[ARG_data12].u_int;
    }
    if (args[ARG_data13].u_int != -1) {
        self->data_pins[13] = args[ARG_data13].u_int;
    }
    if (args[ARG_data14].u_int != -1) {
        self->data_pins[14] = args[ARG_data14].u_int;
    }
    if (args[ARG_data15].u_int != -1) {
        self->data_pins[15] = args[ARG_data15].u_int;
    }
    
    // Set other configuration parameters
    self->freq = args[ARG_freq].u_int;
    self->dc_idle_high = args[ARG_dc_idle_high].u_bool;
    self->dc_cmd_high = args[ARG_dc_cmd_high].u_bool;
    self->dc_dummy_high = args[ARG_dc_dummy_high].u_bool;
    self->dc_data_high = args[ARG_dc_data_high].u_bool;
    self->cs_active_high = args[ARG_cs_active_high].u_bool;
    self->reverse_color_bits = args[ARG_reverse_color_bits].u_bool;
    self->swap_color_bytes = args[ARG_swap_color_bytes].u_bool;
    self->pclk_active_low = args[ARG_pclk_active_low].u_bool;
    self->pclk_idle_low = args[ARG_pclk_idle_low].u_bool;
    
    // Initialize pins, FMC interface
    lcdi8080_init_fmc(self);
    
    return MP_OBJ_FROM_PTR(self);
}

// init method
STATIC mp_obj_t lcdi8080_i80bus_init(size_t n_args, const mp_obj_t *args) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    
    // Parse parameters
    self->width = mp_obj_get_int(args[1]);
    self->height = mp_obj_get_int(args[2]);
    self->bpp = mp_obj_get_int(args[3]);
    self->buffer_size = mp_obj_get_int(args[4]);
    self->rgb565_byte_swap = mp_obj_is_true(args[5]);
    self->cmd_bits = mp_obj_get_int(args[6]);
    self->param_bits = mp_obj_get_int(args[7]);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdi8080_i80bus_init_obj, 8, 8, lcdi8080_i80bus_init);

// deinit method
STATIC mp_obj_t lcdi8080_i80bus_deinit(mp_obj_t self_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Deinitialize FMC and GPIO
    lcdi8080_deinit_fmc(self);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_1(lcdi8080_i80bus_deinit_obj, lcdi8080_i80bus_deinit);

// register_callback method
STATIC mp_obj_t lcdi8080_i80bus_register_callback(mp_obj_t self_in, mp_obj_t callback_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Store callback function
    self->callback = callback_in;
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_2(lcdi8080_i80bus_register_callback_obj, lcdi8080_i80bus_register_callback);

// tx_color method
STATIC mp_obj_t lcdi8080_i80bus_tx_color(size_t n_args, const mp_obj_t *args) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(args[2], &bufinfo, MP_BUFFER_READ);
    
    int start_x = mp_obj_get_int(args[3]);
    int start_y = mp_obj_get_int(args[4]);
    int end_x = mp_obj_get_int(args[5]);
    int end_y = mp_obj_get_int(args[6]);
    int rotation = mp_obj_get_int(args[7]);
    bool last_update = mp_obj_is_true(args[8]);
    
    // Start transmission
    lcdi8080_cs_select(self, true);
    
    // Send command
    lcdi8080_dc_set(self, true, false, false);
    uint8_t cmd_buf[2];
    if (self->cmd_bits > 8) {
        cmd_buf[0] = cmd >> 8;
        cmd_buf[1] = cmd & 0xFF;
        lcdi8080_write_data(self, cmd_buf, 2);
    } else {
        cmd_buf[0] = cmd;
        lcdi8080_write_data(self, cmd_buf, 1);
    }
    
    // Send color data
    lcdi8080_dc_set(self, false, false, true);
    lcdi8080_write_data(self, bufinfo.buf, bufinfo.len);
    
    // End transmission
    lcdi8080_cs_select(self, false);
    
    // Call user callback if registered
    if (self->callback != mp_const_none) {
        mp_obj_t args_callback[9];
        args_callback[0] = self->callback;
        args_callback[1] = MP_OBJ_NEW_SMALL_INT(cmd);
        args_callback[2] = MP_OBJ_FROM_PTR(&bufinfo);
        args_callback[3] = MP_OBJ_NEW_SMALL_INT(start_x);
        args_callback[4] = MP_OBJ_NEW_SMALL_INT(start_y);
        args_callback[5] = MP_OBJ_NEW_SMALL_INT(end_x);
        args_callback[6] = MP_OBJ_NEW_SMALL_INT(end_y);
        args_callback[7] = MP_OBJ_NEW_SMALL_INT(rotation);
        args_callback[8] = mp_obj_new_bool(last_update);
        mp_call_function_n_kw(self->callback, 8, 0, args_callback + 1);
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdi8080_i80bus_tx_color_obj, 9, 9, lcdi8080_i80bus_tx_color);

// tx_param method
STATIC mp_obj_t lcdi8080_i80bus_tx_param(mp_obj_t self_in, mp_obj_t cmd_in, mp_obj_t data_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    uint16_t cmd = mp_obj_get_int(cmd_in);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(data_in, &bufinfo, MP_BUFFER_READ);
    
    // Start transmission
    lcdi8080_cs_select(self, true);
    
    // Send command
    lcdi8080_dc_set(self, true, false, false);
    uint8_t cmd_buf[2];
    if (self->cmd_bits > 8) {
        cmd_buf[0] = cmd >> 8;
        cmd_buf[1] = cmd & 0xFF;
        lcdi8080_write_data(self, cmd_buf, 2);
    } else {
        cmd_buf[0] = cmd;
        lcdi8080_write_data(self, cmd_buf, 1);
    }
    
    // Send parameters
    lcdi8080_dc_set(self, false, false, true);
    lcdi8080_write_data(self, bufinfo.buf, bufinfo.len);
    
    // End transmission
    lcdi8080_cs_select(self, false);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_3(lcdi8080_i80bus_tx_param_obj, lcdi8080_i80bus_tx_param);

// rx_param method
STATIC mp_obj_t lcdi8080_i80bus_rx_param(mp_obj_t self_in, mp_obj_t cmd_in, mp_obj_t params_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    uint16_t cmd = mp_obj_get_int(cmd_in);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(params_in, &bufinfo, MP_BUFFER_WRITE);
    
    // Start transmission
    lcdi8080_cs_select(self, true);
    
    // Send command
    lcdi8080_dc_set(self, true, false, false);
    uint8_t cmd_buf[2];
    if (self->cmd_bits > 8) {
        cmd_buf[0] = cmd >> 8;
        cmd_buf[1] = cmd & 0xFF;
        lcdi8080_write_data(self, cmd_buf, 2);
    } else {
        cmd_buf[0] = cmd;
        lcdi8080_write_data(self, cmd_buf, 1);
    }
    
    // Read parameters
    lcdi8080_dc_set(self, false, false, true);
    lcdi8080_read_data(self, bufinfo.buf, bufinfo.len);
    
    // End transmission
    lcdi8080_cs_select(self, false);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_3(lcdi8080_i80bus_rx_param_obj, lcdi8080_i80bus_rx_param);

// get_lane_count method
STATIC mp_obj_t lcdi8080_i80bus_get_lane_count(mp_obj_t self_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Return the bus width (8 or 16 bits)
    return MP_OBJ_NEW_SMALL_INT(self->data_width / 8);
}
STATIC MP_DEFINE_CONST_FUN_OBJ_1(lcdi8080_i80bus_get_lane_count_obj, lcdi8080_i80bus_get_lane_count);

// allocate_framebuffer method
STATIC mp_obj_t lcdi8080_i80bus_allocate_framebuffer(mp_obj_t self_in, mp_obj_t size_in, mp_obj_t caps_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    size_t size = mp_obj_get_int(size_in);
    uint32_t caps = mp_obj_get_int(caps_in);  // Capabilities flags (not used in this implementation)
    
    // Free existing framebuffer if already allocated
    if (self->framebuffer != NULL) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    // Allocate memory for framebuffer
    self->framebuffer = m_malloc(size);
    if (self->framebuffer == NULL) {
        return mp_const_none;  // Return None if allocation failed
    }
    
    self->framebuffer_size = size;
    
    // Create a memoryview object for the framebuffer
    mp_obj_array_t *mv = m_new_obj(mp_obj_array_t);
    mp_obj_memoryview_init(mv, 'B', 0, size, self->framebuffer);
    
    return MP_OBJ_FROM_PTR(mv);
}
STATIC MP_DEFINE_CONST_FUN_OBJ_3(lcdi8080_i80bus_allocate_framebuffer_obj, lcdi8080_i80bus_allocate_framebuffer);

// free_framebuffer method
STATIC mp_obj_t lcdi8080_i80bus_free_framebuffer(mp_obj_t self_in, mp_obj_t framebuffer_in) {
    lcdi8080_i80bus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Free the framebuffer memory
    if (self->framebuffer != NULL) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_2(lcdi8080_i80bus_free_framebuffer_obj, lcdi8080_i80bus_free_framebuffer);

// Define all methods for the I80Bus class
STATIC const mp_rom_map_elem_t lcdi8080_i80bus_locals_dict_table[] = {
    { MP_ROM_QSTR(MP_QSTR_init), MP_ROM_PTR(&lcdi8080_i80bus_init_obj) },
    { MP_ROM_QSTR(MP_QSTR_deinit), MP_ROM_PTR(&lcdi8080_i80bus_deinit_obj) },
    { MP_ROM_QSTR(MP_QSTR_register_callback), MP_ROM_PTR(&lcdi8080_i80bus_register_callback_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_color), MP_ROM_PTR(&lcdi8080_i80bus_tx_color_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_param), MP_ROM_PTR(&lcdi8080_i80bus_tx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_rx_param), MP_ROM_PTR(&lcdi8080_i80bus_rx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_get_lane_count), MP_ROM_PTR(&lcdi8080_i80bus_get_lane_count_obj) },
    { MP_ROM_QSTR(MP_QSTR_allocate_framebuffer), MP_ROM_PTR(&lcdi8080_i80bus_allocate_framebuffer_obj) },
    { MP_ROM_QSTR(MP_QSTR_free_framebuffer), MP_ROM_PTR(&lcdi8080_i80bus_free_framebuffer_obj) },
};
STATIC MP_DEFINE_CONST_DICT(lcdi8080_i80bus_locals_dict, lcdi8080_i80bus_locals_dict_table);

// Define the I80Bus type
const mp_obj_type_t lcdi8080_i80bus_type = {
    { &mp_type_type },
    .name = MP_QSTR_I80Bus,
    .make_new = lcdi8080_i80bus_make_new,
    .locals_dict = (mp_obj_dict_t *)&lcdi8080_i80bus_locals_dict,
};

// Define module globals
STATIC const mp_rom_map_elem_t lcdi8080_module_globals_table[] = {
    { MP_ROM_QSTR(MP_QSTR___name__), MP_ROM_QSTR(MP_QSTR_lcdi8080) },
    { MP_ROM_QSTR(MP_QSTR_I80Bus), MP_ROM_PTR(&lcdi8080_i80bus_type) },
};
STATIC MP_DEFINE_CONST_DICT(lcdi8080_module_globals, lcdi8080_module_globals_table);

// Define the module object
const mp_obj_module_t lcdi8080_user_cmodule = {
    .base = { &mp_type_module },
    .globals = (mp_obj_dict_t *)&lcdi8080_module_globals,
};

// Register the module to make it available in Python
MP_REGISTER_MODULE(MP_QSTR_lcdi8080, lcdi8080_user_cmodule);