#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "lcdrgb.h"

// STM32 specific includes
#include "stm32f4xx_hal.h"

// LTDC handle
static LTDC_HandleTypeDef hltdc = {0};

// Initialize LTDC and GPIO pins
static void lcdrgb_init_ltdc(lcdrgb_rgbbus_obj_t *self) {
    // Enable LTDC clock
    __HAL_RCC_LTDC_CLK_ENABLE();
    
    // Enable GPIO clocks
    __HAL_RCC_GPIOA_CLK_ENABLE();
    __HAL_RCC_GPIOB_CLK_ENABLE();
    __HAL_RCC_GPIOC_CLK_ENABLE();
    __HAL_RCC_GPIOD_CLK_ENABLE();
    __HAL_RCC_GPIOE_CLK_ENABLE();
    __HAL_RCC_GPIOF_CLK_ENABLE();
    __HAL_RCC_GPIOG_CLK_ENABLE();
    
    // Configure GPIO pins for LTDC signals
    GPIO_InitTypeDef GPIO_InitStruct = {0};
    
    // Common configuration for all LCD pins
    GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
    GPIO_InitStruct.Pull = GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_HIGH;
    GPIO_InitStruct.Alternate = GPIO_AF14_LTDC;  // Most LTDC pins use AF14
    
    // Configure control pins
    // This is a simplified implementation - in a real driver you'd need to determine
    // which GPIO port each pin is connected to
    
    // For demonstration purposes, we'll assume specific pin mappings
    // HSYNC pin
    GPIO_InitStruct.Pin = GPIO_PIN_0;  // Example pin
    HAL_GPIO_Init(GPIOC, &GPIO_InitStruct);
    
    // VSYNC pin
    GPIO_InitStruct.Pin = GPIO_PIN_1;  // Example pin
    HAL_GPIO_Init(GPIOC, &GPIO_InitStruct);
    
    // DE pin
    GPIO_InitStruct.Pin = GPIO_PIN_2;  // Example pin
    HAL_GPIO_Init(GPIOC, &GPIO_InitStruct);
    
    // PCLK pin
    GPIO_InitStruct.Pin = GPIO_PIN_3;  // Example pin
    HAL_GPIO_Init(GPIOC, &GPIO_InitStruct);
    
    // Configure data pins
    // In a real driver, you'd need to configure each pin individually based on their
    // GPIO port and pin number
    // Example for data pins 0-7:
    GPIO_InitStruct.Pin = 0xFF;  // Bits 0-7
    HAL_GPIO_Init(GPIOD, &GPIO_InitStruct);
    
    // Initialize LTDC
    self->ltdc_handle = &hltdc;
    LTDC_HandleTypeDef *hltdc_ptr = (LTDC_HandleTypeDef *)self->ltdc_handle;
    
    // Configure LTDC timing parameters
    hltdc_ptr->Instance = LTDC;
    hltdc_ptr->Init.HSPolarity = self->hsync_idle_low ? LTDC_HSPOLARITY_AL : LTDC_HSPOLARITY_AH;
    hltdc_ptr->Init.VSPolarity = self->vsync_idle_low ? LTDC_VSPOLARITY_AL : LTDC_VSPOLARITY_AH;
    hltdc_ptr->Init.DEPolarity = self->de_idle_high ? LTDC_DEPOLARITY_AH : LTDC_DEPOLARITY_AL;
    hltdc_ptr->Init.PCPolarity = self->pclk_active_low ? LTDC_PCPOLARITY_IPC : LTDC_PCPOLARITY_IIPC;
    
    // Horizontal configuration
    hltdc_ptr->Init.HorizontalSync = self->hsync_pulse_width - 1;
    hltdc_ptr->Init.AccumulatedHBP = self->hsync_pulse_width + self->hsync_back_porch - 1;
    hltdc_ptr->Init.AccumulatedActiveW = self->hsync_pulse_width + self->hsync_back_porch + self->width - 1;
    hltdc_ptr->Init.TotalWidth = self->hsync_pulse_width + self->hsync_back_porch + self->width + self->hsync_front_porch - 1;
    
    // Vertical configuration
    hltdc_ptr->Init.VerticalSync = self->vsync_pulse_width - 1;
    hltdc_ptr->Init.AccumulatedVBP = self->vsync_pulse_width + self->vsync_back_porch - 1;
    hltdc_ptr->Init.AccumulatedActiveH = self->vsync_pulse_width + self->vsync_back_porch + self->height - 1;
    hltdc_ptr->Init.TotalHeigh = self->vsync_pulse_width + self->vsync_back_porch + self->height + self->vsync_front_porch - 1;
    
    // Background configuration
    hltdc_ptr->Init.Backcolor.Blue = 0;
    hltdc_ptr->Init.Backcolor.Green = 0;
    hltdc_ptr->Init.Backcolor.Red = 0;
    
    // Initialize LTDC
    if (HAL_LTDC_Init(hltdc_ptr) != HAL_OK) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("Failed to initialize LTDC"));
    }
    
    // Configure layer
    LTDC_LayerCfgTypeDef pLayerCfg = {0};
    
    // Common configuration for both layers
    pLayerCfg.WindowX0 = 0;
    pLayerCfg.WindowX1 = self->width;
    pLayerCfg.WindowY0 = 0;
    pLayerCfg.WindowY1 = self->height;
    pLayerCfg.Alpha = 255;
    pLayerCfg.Alpha0 = 0;
    pLayerCfg.BlendingFactor1 = LTDC_BLENDING_FACTOR1_CA;
    pLayerCfg.BlendingFactor2 = LTDC_BLENDING_FACTOR2_CA;
    
    // Set pixel format based on bpp
    if (self->bpp == 16) {
        pLayerCfg.PixelFormat = LTDC_PIXEL_FORMAT_RGB565;
    } else if (self->bpp == 24) {
        pLayerCfg.PixelFormat = LTDC_PIXEL_FORMAT_RGB888;
    } else {
        mp_raise_ValueError(MP_ERROR_TEXT("Unsupported BPP"));
    }
    
    // Configure default color for transparency
    pLayerCfg.Backcolor.Blue = 0;
    pLayerCfg.Backcolor.Green = 0;
    pLayerCfg.Backcolor.Red = 0;
    
    // Configure the layer
    if (HAL_LTDC_ConfigLayer(hltdc_ptr, &pLayerCfg, 0) != HAL_OK) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("Failed to configure LTDC layer"));
    }
}

// Deinitialize LTDC and GPIO
static void lcdrgb_deinit_ltdc(lcdrgb_rgbbus_obj_t *self) {
    LTDC_HandleTypeDef *hltdc_ptr = (LTDC_HandleTypeDef *)self->ltdc_handle;
    
    // Deinitialize LTDC
    if (hltdc_ptr != NULL) {
        HAL_LTDC_DeInit(hltdc_ptr);
    }
    
    // Deinitialize GPIO pins (simplified - in real driver, deinitialize all pins)
    HAL_GPIO_DeInit(GPIOC, GPIO_PIN_0); // HSYNC
    HAL_GPIO_DeInit(GPIOC, GPIO_PIN_1); // VSYNC
    HAL_GPIO_DeInit(GPIOC, GPIO_PIN_2); // DE
    HAL_GPIO_DeInit(GPIOC, GPIO_PIN_3); // PCLK
    HAL_GPIO_DeInit(GPIOD, 0xFF);       // Data pins 0-7
}

// Update display content (copy framebuffer to LTDC layer)
static void lcdrgb_update_display(lcdrgb_rgbbus_obj_t *self) {
    LTDC_HandleTypeDef *hltdc_ptr = (LTDC_HandleTypeDef *)self->ltdc_handle;
    
    // Set framebuffer address for layer 0
    if (self->framebuffer != NULL) {
        HAL_LTDC_SetAddress(hltdc_ptr, (uint32_t)self->framebuffer, 0);
    }
}