#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "py/mperrno.h"
#include "py/objarray.h"

#include "lcdrgb.h"

// STM32 specific includes
#include "stm32f4xx_hal.h"

// Helper function declarations
static void lcdrgb_init_ltdc(lcdrgb_rgbbus_obj_t *self);
static void lcdrgb_deinit_ltdc(lcdrgb_rgbbus_obj_t *self);
static void lcdrgb_update_display(lcdrgb_rgbbus_obj_t *self);

// RGBBus constructor
STATIC mp_obj_t lcdrgb_rgbbus_make_new(const mp_obj_type_t *type, size_t n_args, size_t n_kw, const mp_obj_t *all_args) {
    enum { 
        ARG_hsync, ARG_vsync, ARG_de, ARG_pclk, 
        ARG_data0, ARG_data1, ARG_data2, ARG_data3, ARG_data4, ARG_data5, ARG_data6, ARG_data7,
        ARG_data8, ARG_data9, ARG_data10, ARG_data11, ARG_data12, ARG_data13, ARG_data14, ARG_data15,
        ARG_freq, ARG_hsync_front_porch, ARG_hsync_back_porch, ARG_hsync_pulse_width, ARG_hsync_idle_low,
        ARG_vsync_front_porch, ARG_vsync_back_porch, ARG_vsync_pulse_width, ARG_vsync_idle_low,
        ARG_de_idle_high, ARG_pclk_idle_high, ARG_pclk_active_low, ARG_disp_active_low,
        ARG_refresh_on_demand, ARG_rgb565_dither
    };
    
    static const mp_arg_t allowed_args[] = {
        { MP_QSTR_hsync, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_vsync, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_de, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_pclk, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data0, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data1, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data2, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data3, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data4, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data5, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data6, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data7, MP_ARG_KW_ONLY | MP_ARG_REQUIRED | MP_ARG_INT },
        { MP_QSTR_data8, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data9, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data10, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data11, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data12, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data13, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data14, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_data15, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = -1} },
        { MP_QSTR_freq, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 8000000} },
        { MP_QSTR_hsync_front_porch, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 0} },
        { MP_QSTR_hsync_back_porch, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 0} },
        { MP_QSTR_hsync_pulse_width, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 1} },
        { MP_QSTR_hsync_idle_low, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_vsync_front_porch, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 0} },
        { MP_QSTR_vsync_back_porch, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 0} },
        { MP_QSTR_vsync_pulse_width, MP_ARG_KW_ONLY | MP_ARG_INT, {.u_int = 1} },
        { MP_QSTR_vsync_idle_low, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_de_idle_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_pclk_idle_high, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_pclk_active_low, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_disp_active_low, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_refresh_on_demand, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
        { MP_QSTR_rgb565_dither, MP_ARG_KW_ONLY | MP_ARG_BOOL, {.u_bool = false} },
    };
    
    mp_arg_val_t args[MP_ARRAY_SIZE(allowed_args)];
    mp_arg_parse_all_kw_array(n_args, n_kw, all_args, MP_ARRAY_SIZE(allowed_args), allowed_args, args);
    
    // Create new object
    lcdrgb_rgbbus_obj_t *self = mp_obj_malloc(lcdrgb_rgbbus_obj_t, &lcdrgb_rgbbus_type);
    
    // Set control pins
    self->hsync_pin = args[ARG_hsync].u_int;
    self->vsync_pin = args[ARG_vsync].u_int;
    self->de_pin = args[ARG_de].u_int;
    self->pclk_pin = args[ARG_pclk].u_int;
    
    // Set data pins
    self->data_pins[0] = args[ARG_data0].u_int;
    self->data_pins[1] = args[ARG_data1].u_int;
    self->data_pins[2] = args[ARG_data2].u_int;
    self->data_pins[3] = args[ARG_data3].u_int;
    self->data_pins[4] = args[ARG_data4].u_int;
    self->data_pins[5] = args[ARG_data5].u_int;
    self->data_pins[6] = args[ARG_data6].u_int;
    self->data_pins[7] = args[ARG_data7].u_int;
    self->data_pins[8] = args[ARG_data8].u_int;
    self->data_pins[9] = args[ARG_data9].u_int;
    self->data_pins[10] = args[ARG_data10].u_int;
    self->data_pins[11] = args[ARG_data11].u_int;
    self->data_pins[12] = args[ARG_data12].u_int;
    self->data_pins[13] = args[ARG_data13].u_int;
    self->data_pins[14] = args[ARG_data14].u_int;
    self->data_pins[15] = args[ARG_data15].u_int;
    
    // Calculate lane count (number of data lines)
    self->lane_count = 8; // Start with minimum 8 bits
    for (int i = 8; i < 16; i++) {
        if (self->data_pins[i] != -1) {
            self->lane_count++;
        }
    }
    
    // Set timing parameters
    self->freq = args[ARG_freq].u_int;
    self->hsync_front_porch = args[ARG_hsync_front_porch].u_int;
    self->hsync_back_porch = args[ARG_hsync_back_porch].u_int;
    self->hsync_pulse_width = args[ARG_hsync_pulse_width].u_int;
    self->hsync_idle_low = args[ARG_hsync_idle_low].u_bool;
    self->vsync_front_porch = args[ARG_vsync_front_porch].u_int;
    self->vsync_back_porch = args[ARG_vsync_back_porch].u_int;
    self->vsync_pulse_width = args[ARG_vsync_pulse_width].u_int;
    self->vsync_idle_low = args[ARG_vsync_idle_low].u_bool;
    self->de_idle_high = args[ARG_de_idle_high].u_bool;
    self->pclk_idle_high = args[ARG_pclk_idle_high].u_bool;
    self->pclk_active_low = args[ARG_pclk_active_low].u_bool;
    self->disp_active_low = args[ARG_disp_active_low].u_bool;
    self->refresh_on_demand = args[ARG_refresh_on_demand].u_bool;
    self->rgb565_dither = args[ARG_rgb565_dither].u_bool;
    
    // Initialize other members
    self->framebuffer = NULL;
    self->framebuffer_size = 0;
    
    return MP_OBJ_FROM_PTR(self);
}

// init method
STATIC mp_obj_t lcdrgb_rgbbus_init(size_t n_args, const mp_obj_t *args) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    
    // Parse parameters
    self->width = mp_obj_get_int(args[1]);
    self->height = mp_obj_get_int(args[2]);
    self->bpp = mp_obj_get_int(args[3]);
    self->buffer_size = mp_obj_get_int(args[4]);
    self->rgb565_byte_swap = mp_obj_is_true(args[5]);
    self->cmd_bits = mp_obj_get_int(args[6]);
    self->param_bits = mp_obj_get_int(args[7]);
    
    // Initialize LTDC
    lcdrgb_init_ltdc(self);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdrgb_rgbbus_init_obj, 8, 8, lcdrgb_rgbbus_init);

// deinit method
STATIC mp_obj_t lcdrgb_rgbbus_deinit(mp_obj_t self_in) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Deinitialize LTDC
    lcdrgb_deinit_ltdc(self);
    
    // Free framebuffer if allocated
    if (self->framebuffer != NULL) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_1(lcdrgb_rgbbus_deinit_obj, lcdrgb_rgbbus_deinit);

// register_callback method
STATIC mp_obj_t lcdrgb_rgbbus_register_callback(mp_obj_t self_in, mp_obj_t callback_in) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Store callback function
    self->callback = callback_in;
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_2(lcdrgb_rgbbus_register_callback_obj, lcdrgb_rgbbus_register_callback);

// tx_param method
STATIC mp_obj_t lcdrgb_rgbbus_tx_param(size_t n_args, const mp_obj_t *args) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    
    // In a pure LTDC implementation, there's typically no direct command
    // transmission. This method might be used for external controllers
    // connected via GPIO or other means.
    
    // For now, we'll just store the command or take some action based on it
    // As a placeholder, we could update the display if a certain command is received
    if (cmd == 0x2C) { // Memory write command in common LCD controllers
        lcdrgb_update_display(self);
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdrgb_rgbbus_tx_param_obj, 2, 3, lcdrgb_rgbbus_tx_param);

// tx_color method - write color data to framebuffer
STATIC mp_obj_t lcdrgb_rgbbus_tx_color(size_t n_args, const mp_obj_t *args) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(args[2], &bufinfo, MP_BUFFER_READ);
    
    int start_x = mp_obj_get_int(args[3]);
    int start_y = mp_obj_get_int(args[4]);
    int end_x = mp_obj_get_int(args[5]);
    int end_y = mp_obj_get_int(args[6]);
    int rotation = mp_obj_get_int(args[7]);
    bool last_update = mp_obj_is_true(args[8]);
    
    // Check if framebuffer is allocated
    if (self->framebuffer == NULL) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("Framebuffer not allocated"));
    }
    
    int width = end_x - start_x + 1;
    int height = end_y - start_y + 1;
    
    // Copy data to framebuffer
    // This is a simplified implementation - in real code, you'd handle rotation
    // and different pixel formats correctly
    
    if (self->bpp == 16) {
        // RGB565 format
        uint16_t *dst = (uint16_t *)self->framebuffer + (start_y * self->width + start_x);
        uint16_t *src = (uint16_t *)bufinfo.buf;
        
        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                // Handle byte swapping if needed
                if (self->rgb565_byte_swap) {
                    uint16_t pixel = src[y * width + x];
                    dst[y * self->width + x] = (pixel << 8) | (pixel >> 8);
                } else {
                    dst[y * self->width + x] = src[y * width + x];
                }
            }
        }
    } else if (self->bpp == 24) {
        // RGB888 format - 3 bytes per pixel
        uint8_t *dst = (uint8_t *)self->framebuffer + ((start_y * self->width + start_x) * 3);
        uint8_t *src = (uint8_t *)bufinfo.buf;
        
        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                int src_idx = (y * width + x) * 3;
                int dst_idx = y * self->width * 3 + x * 3;
                
                dst[dst_idx] = src[src_idx];       // R
                dst[dst_idx + 1] = src[src_idx + 1]; // G
                dst[dst_idx + 2] = src[src_idx + 2]; // B
            }
        }
    }
    
    // Update display if this is the last update or refresh on demand is disabled
    if (last_update || !self->refresh_on_demand) {
        lcdrgb_update_display(self);
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdrgb_rgbbus_tx_color_obj, 9, 9, lcdrgb_rgbbus_tx_color);

// rx_param method
STATIC mp_obj_t lcdrgb_rgbbus_rx_param(size_t n_args, const mp_obj_t *args) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(args[0]);
    uint8_t cmd = mp_obj_get_int(args[1]);
    
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(args[2], &bufinfo, MP_BUFFER_WRITE);
    
    // In a pure LTDC implementation, there's typically no direct command
    // reading. This method might be used for external controllers
    // connected via GPIO or other means.
    
    // For demonstration purposes, we'll just fill the buffer with a default value
    // In a real implementation, you would read from the display controller
    memset(bufinfo.buf, 0, bufinfo.len);
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(lcdrgb_rgbbus_rx_param_obj, 3, 3, lcdrgb_rgbbus_rx_param);

// get_lane_count method
STATIC mp_obj_t lcdrgb_rgbbus_get_lane_count(mp_obj_t self_in) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    return mp_obj_new_int(self->lane_count);
}
STATIC MP_DEFINE_CONST_FUN_OBJ_1(lcdrgb_rgbbus_get_lane_count_obj, lcdrgb_rgbbus_get_lane_count);

// allocate_framebuffer method
STATIC mp_obj_t lcdrgb_rgbbus_allocate_framebuffer(mp_obj_t self_in, mp_obj_t size_in, mp_obj_t caps_in) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    size_t size = mp_obj_get_int(size_in);
    int caps = mp_obj_get_int(caps_in);  // Memory capabilities flags
    
    // Free existing framebuffer if any
    if (self->framebuffer != NULL) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    }
    
    // Allocate new framebuffer
    // For STM32, we need to ensure proper alignment and location
    // In a real implementation, you might use different allocation strategies
    // based on the 'caps' parameter
    
    // Align to 32 bytes for DMA transfers
    size = (size + 31) & ~31;
    
    void *buffer = m_malloc(size);
    if (buffer == NULL) {
        // Return None if allocation failed
        return mp_const_none;
    }
    
    // Store framebuffer pointer and size
    self->framebuffer = buffer;
    self->framebuffer_size = size;
    
    // Clear the framebuffer
    memset(buffer, 0, size);
    
    // Create and return a memoryview object for the buffer
    mp_obj_array_t *memview = m_new_obj(mp_obj_array_t);
    memview->base.type = &mp_type_memoryview;
    memview->typecode = 'B';  // unsigned char
    memview->free = 0;        // don't free buffer when memoryview is freed
    memview->len = size;
    memview->items = buffer;
    
    return MP_OBJ_FROM_PTR(memview);
}
STATIC MP_DEFINE_CONST_FUN_OBJ_3(lcdrgb_rgbbus_allocate_framebuffer_obj, lcdrgb_rgbbus_allocate_framebuffer);

// free_framebuffer method
STATIC mp_obj_t lcdrgb_rgbbus_free_framebuffer(mp_obj_t self_in, mp_obj_t framebuffer_in) {
    lcdrgb_rgbbus_obj_t *self = MP_OBJ_TO_PTR(self_in);
    
    // Get buffer info from memoryview object
    mp_buffer_info_t bufinfo;
    mp_get_buffer_raise(framebuffer_in, &bufinfo, MP_BUFFER_READ);
    
    // Check if this is our framebuffer
    if (bufinfo.buf == self->framebuffer) {
        m_free(self->framebuffer);
        self->framebuffer = NULL;
        self->framebuffer_size = 0;
    } else {
        mp_raise_ValueError(MP_ERROR_TEXT("Not the allocated framebuffer"));
    }
    
    return mp_const_none;
}
STATIC MP_DEFINE_CONST_FUN_OBJ_2(lcdrgb_rgbbus_free_framebuffer_obj, lcdrgb_rgbbus_free_framebuffer);

// Define all methods for the RGBBus class
STATIC const mp_rom_map_elem_t lcdrgb_rgbbus_locals_dict_table[] = {
    { MP_ROM_QSTR(MP_QSTR_init), MP_ROM_PTR(&lcdrgb_rgbbus_init_obj) },
    { MP_ROM_QSTR(MP_QSTR_deinit), MP_ROM_PTR(&lcdrgb_rgbbus_deinit_obj) },
    { MP_ROM_QSTR(MP_QSTR_register_callback), MP_ROM_PTR(&lcdrgb_rgbbus_register_callback_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_param), MP_ROM_PTR(&lcdrgb_rgbbus_tx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_tx_color), MP_ROM_PTR(&lcdrgb_rgbbus_tx_color_obj) },
    { MP_ROM_QSTR(MP_QSTR_rx_param), MP_ROM_PTR(&lcdrgb_rgbbus_rx_param_obj) },
    { MP_ROM_QSTR(MP_QSTR_get_lane_count), MP_ROM_PTR(&lcdrgb_rgbbus_get_lane_count_obj) },
    { MP_ROM_QSTR(MP_QSTR_allocate_framebuffer), MP_ROM_PTR(&lcdrgb_rgbbus_allocate_framebuffer_obj) },
    { MP_ROM_QSTR(MP_QSTR_free_framebuffer), MP_ROM_PTR(&lcdrgb_rgbbus_free_framebuffer_obj) },
};
STATIC MP_DEFINE_CONST_DICT(lcdrgb_rgbbus_locals_dict, lcdrgb_rgbbus_locals_dict_table);

// Define the RGBBus type
const mp_obj_type_t lcdrgb_rgbbus_type = {
    { &mp_type_type },
    .name = MP_QSTR_RGBBus,
    .make_new = lcdrgb_rgbbus_make_new,
    .locals_dict = (mp_obj_dict_t *)&lcdrgb_rgbbus_locals_dict,
};

// Module level functions

// Define the module globals table
STATIC const mp_rom_map_elem_t lcdrgb_module_globals_table[] = {
    { MP_ROM_QSTR(MP_QSTR___name__), MP_ROM_QSTR(MP_QSTR_lcdrgb) },
    { MP_ROM_QSTR(MP_QSTR_RGBBus), MP_ROM_PTR(&lcdrgb_rgbbus_type) },
};
STATIC MP_DEFINE_CONST_DICT(lcdrgb_module_globals, lcdrgb_module_globals_table);

// Define the module object
const mp_obj_module_t lcdrgb_user_cmodule = {
    .base = { &mp_type_module },
    .globals = (mp_obj_dict_t *)&lcdrgb_module_globals,
};

// Register the module
MP_REGISTER_MODULE(MP_QSTR_lcdrgb, lcdrgb_user_cmodule);