#include <stdio.h>
#include <string.h>

#include "py/runtime.h"
#include "py/mphal.h"
#include "lcdspi.h"

// STM32 specific includes
#include "stm32f4xx_hal.h"

// SPI handles array (one for each SPI peripheral)
static SPI_HandleTypeDef spi_handles[6] = {0};

// Initialize SPI and GPIO pins
static void lcdspi_init_spi(lcdspi_spibus_obj_t *self) {
    // Configure GPIO pins for DC and CS
    GPIO_InitTypeDef GPIO_InitStruct = {0};
    
    // Enable GPIO clocks (this assumes the pins are on GPIOA, adjust as needed)
    __HAL_RCC_GPIOA_CLK_ENABLE();
    
    // Configure DC pin
    GPIO_InitStruct.Pin = (1 << self->dc_pin);
    GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
    GPIO_InitStruct.Pull = GPIO_NOPULL;
    GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_HIGH;
    HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);
    
    // Configure CS pin if used
    if (self->cs_pin >= 0) {
        GPIO_InitStruct.Pin = (1 << self->cs_pin);
        HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);
        
        // Initial CS state (inactive)
        HAL_GPIO_WritePin(GPIOA, (1 << self->cs_pin), 
                         self->cs_high_active ? GPIO_PIN_RESET : GPIO_PIN_SET);
    }
    
    // Configure and initialize SPI
    SPI_HandleTypeDef *hspi = &spi_handles[self->spi_bus];
    self->spi_handle = hspi;
    
    // Enable SPI clock based on the bus number
    switch (self->spi_bus) {
        case 1:
            __HAL_RCC_SPI1_CLK_ENABLE();
            hspi->Instance = SPI1;
            break;
        case 2:
            __HAL_RCC_SPI2_CLK_ENABLE();
            hspi->Instance = SPI2;
            break;
        case 3:
            __HAL_RCC_SPI3_CLK_ENABLE();
            hspi->Instance = SPI3;
            break;
        // Add more cases for other SPI peripherals if needed
        default:
            mp_raise_ValueError(MP_ERROR_TEXT("Invalid SPI bus"));
    }
    
    // Configure SPI parameters
    hspi->Init.Mode = SPI_MODE_MASTER;
    
    // Set SPI mode (CPOL and CPHA)
    switch (self->spi_mode) {
        case 0: // CPOL=0, CPHA=0
            hspi->Init.CLKPolarity = SPI_POLARITY_LOW;
            hspi->Init.CLKPhase = SPI_PHASE_1EDGE;
            break;
        case 1: // CPOL=0, CPHA=1
            hspi->Init.CLKPolarity = SPI_POLARITY_LOW;
            hspi->Init.CLKPhase = SPI_PHASE_2EDGE;
            break;
        case 2: // CPOL=1, CPHA=0
            hspi->Init.CLKPolarity = SPI_POLARITY_HIGH;
            hspi->Init.CLKPhase = SPI_PHASE_1EDGE;
            break;
        case 3: // CPOL=1, CPHA=1
            hspi->Init.CLKPolarity = SPI_POLARITY_HIGH;
            hspi->Init.CLKPhase = SPI_PHASE_2EDGE;
            break;
    }
    
    // Set bit order
    hspi->Init.FirstBit = self->lsb_first ? SPI_FIRSTBIT_LSB : SPI_FIRSTBIT_MSB;
    
    // Calculate prescaler for frequency
    uint32_t apb_clock = HAL_RCC_GetPCLK1Freq();
    if (self->spi_bus == 1) {
        apb_clock = HAL_RCC_GetPCLK2Freq(); // SPI1 is on APB2
    }
    
    // Calculate the prescaler for the desired frequency
    uint32_t prescaler = apb_clock / self->freq;
    uint32_t prescaler_bits;
    
    if (prescaler <= 2) prescaler_bits = SPI_BAUDRATEPRESCALER_2;
    else if (prescaler <= 4) prescaler_bits = SPI_BAUDRATEPRESCALER_4;
    else if (prescaler <= 8) prescaler_bits = SPI_BAUDRATEPRESCALER_8;
    else if (prescaler <= 16) prescaler_bits = SPI_BAUDRATEPRESCALER_16;
    else if (prescaler <= 32) prescaler_bits = SPI_BAUDRATEPRESCALER_32;
    else if (prescaler <= 64) prescaler_bits = SPI_BAUDRATEPRESCALER_64;
    else if (prescaler <= 128) prescaler_bits = SPI_BAUDRATEPRESCALER_128;
    else prescaler_bits = SPI_BAUDRATEPRESCALER_256;
    
    hspi->Init.BaudRatePrescaler = prescaler_bits;
    hspi->Init.Direction = SPI_DIRECTION_2LINES;
    hspi->Init.DataSize = SPI_DATASIZE_8BIT;
    hspi->Init.NSS = SPI_NSS_SOFT;
    hspi->Init.TIMode = SPI_TIMODE_DISABLE;
    hspi->Init.CRCCalculation = SPI_CRCCALCULATION_DISABLE;
    hspi->Init.CRCPolynomial = 10;
    
    // Initialize SPI
    if (HAL_SPI_Init(hspi) != HAL_OK) {
        mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("Failed to initialize SPI"));
    }
}

// Deinitialize SPI and GPIO
static void lcdspi_deinit_spi(lcdspi_spibus_obj_t *self) {
    SPI_HandleTypeDef *hspi = (SPI_HandleTypeDef *)self->spi_handle;
    
    // Deinitialize SPI
    if (hspi != NULL) {
        HAL_SPI_DeInit(hspi);
    }
    
    // Deinitialize GPIO pins
    if (self->dc_pin >= 0) {
        HAL_GPIO_DeInit(GPIOA, (1 << self->dc_pin));
    }
    
    if (self->cs_pin >= 0) {
        HAL_GPIO_DeInit(GPIOA, (1 << self->cs_pin));
    }
}

// Select/deselect chip using CS pin
static void lcdspi_cs_select(lcdspi_spibus_obj_t *self, bool select) {
    if (self->cs_pin >= 0) {
        bool pin_value;
        
        if (self->cs_high_active) {
            pin_value = select ? GPIO_PIN_SET : GPIO_PIN_RESET;
        } else {
            pin_value = select ? GPIO_PIN_RESET : GPIO_PIN_SET;
        }
        
        HAL_GPIO_WritePin(GPIOA, (1 << self->cs_pin), pin_value);
    }
}

// Set DC pin for command or data mode
static void lcdspi_dc_select(lcdspi_spibus_obj_t *self, bool cmd) {
    bool pin_value;
    
    if (self->dc_low_on_data) {
        pin_value = cmd ? GPIO_PIN_SET : GPIO_PIN_RESET;
    } else {
        pin_value = cmd ? GPIO_PIN_RESET : GPIO_PIN_SET;
    }
    
    HAL_GPIO_WritePin(GPIOA, (1 << self->dc_pin), pin_value);
}

// Send data over SPI
static void lcdspi_send_data(lcdspi_spibus_obj_t *self, const uint8_t *data, size_t len) {
    SPI_HandleTypeDef *hspi = (SPI_HandleTypeDef *)self->spi_handle;
    
    // Split into chunks if needed (HAL_SPI_Transmit may have limitations)
    const size_t chunk_size = 1024;
    
    for (size_t i = 0; i < len; i += chunk_size) {
        size_t send_len = (i + chunk_size > len) ? (len - i) : chunk_size;
        
        if (HAL_SPI_Transmit(hspi, (uint8_t *)data + i, send_len, HAL_MAX_DELAY) != HAL_OK) {
            mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("SPI transmission failed"));
        }
    }
}

// Receive data over SPI
static void lcdspi_receive_data(lcdspi_spibus_obj_t *self, uint8_t *data, size_t len) {
    SPI_HandleTypeDef *hspi = (SPI_HandleTypeDef *)self->spi_handle;
    
    // Split into chunks if needed
    const size_t chunk_size = 1024;
    
    for (size_t i = 0; i < len; i += chunk_size) {
        size_t recv_len = (i + chunk_size > len) ? (len - i) : chunk_size;
        
        if (HAL_SPI_Receive(hspi, data + i, recv_len, HAL_MAX_DELAY) != HAL_OK) {
            mp_raise_msg(&mp_type_OSError, MP_ERROR_TEXT("SPI reception failed"));
        }
    }
}